#include "mapeditor.h"

#include <q3dragobject.h>
#include <q3urloperator.h>
#include <q3networkprotocol.h>
#include <q3paintdevicemetrics.h>
#include <q3filedialog.h>

#include <iostream>
#include <cstdlib>
#include <typeinfo>

#include "version.h"

#include "api.h"
#include "editxlinkdialog.h"
#include "exports.h"
#include "extrainfodialog.h"
#include "file.h"
#include "linkablemapobj.h"
#include "mainwindow.h"
#include "misc.h"
#include "texteditor.h"
#include "warningdialog.h"
#include "xml.h"


extern TextEditor *textEditor;
extern int statusbarTime;
extern Main *mainWindow;
extern QString tmpVymDir;
extern QString clipboardDir;
extern bool clipboardEmpty;
extern FlagRowObj *standardFlagsDefault;

extern QMenu* branchContextMenu;
extern QMenu* branchAddContextMenu;
extern QMenu* branchRemoveContextMenu;
extern QMenu* branchLinksContextMenu;
extern QMenu* branchXLinksContextMenuEdit;
extern QMenu* branchXLinksContextMenuFollow;
extern QMenu* floatimageContextMenu;
extern QMenu* canvasContextMenu;


extern Settings settings;
extern ImageIO imageIO;

extern QString vymName;
extern QString vymVersion;

extern QString iconPath;
extern QDir vymBaseDir;
extern QDir lastImageDir;

int MapEditor::mapNum=0;	// make instance

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
MapEditor::MapEditor(
	QWidget* parent, const char* name, Qt::WFlags f) :
  Q3CanvasView(parent,name,f), urlOperator(0), imageBuffer(0)
{
	//cout << "Constructor ME "<<this<<endl;
	mapNum++;

    viewport()->setAcceptDrops(true);

    mapCanvas = new Q3Canvas(width(),height());
	mapCanvas->setAdvancePeriod(30);
	mapCanvas->setBackgroundColor (Qt::white);

    setCanvas (mapCanvas);
	
	// Always show scroll bars (automatic would flicker sometimes)
	setVScrollBarMode ( Q3ScrollView::AlwaysOn );
	setHScrollBarMode ( Q3ScrollView::AlwaysOn );

    mapCenter = new MapCenterObj(mapCanvas);
    mapCenter->setVisibility (true);
	mapCenter->setMapEditor (this);
	mapCenter->setHeading (tr("New Map","Heading of mapcenter in new map"));
	mapCenter->move(mapCanvas->width()/2-mapCenter->width()/2,mapCanvas->height()/2-mapCenter->height()/2);

    printer=NULL;

	defLinkColor=QColor (0,0,255);
	defXLinkColor=QColor (180,180,180);
	linkcolorhint=DefaultColor;
	linkstyle=StylePolyParabel;

	// Create bitmap cursors, platform dependant
	HandOpenCursor=QCursor (QPixmap(iconPath+"cursorhandopen.png"),1,1);		
	PickColorCursor=QCursor ( QPixmap(iconPath+"cursorcolorpicker.png"), 5,27 ); 
	CopyCursor=QCursor ( QPixmap(iconPath+"cursorcopy.png"), 1,1 ); 
	XLinkCursor=QCursor ( QPixmap(iconPath+"cursorxlink.png"), 1,7 ); 

	setFocusPolicy (Qt::StrongFocus);

	pickingColor=false;
	drawingLink=false;
	copyingObj=false;

    editingBO=NULL;
    selection=NULL;
    selectionLast=NULL;
    movingObj=NULL;

	xelection.setMapCenter (mapCenter);

	defXLinkWidth=1;
	defXLinkColor=QColor (230,230,230);

    mapChanged=false;
	mapDefault=true;
	mapUnsaved=false;
	
	zipped=true;
	filePath="";
	fileName=tr("unnamed");
	mapName="";

	stepsTotal=settings.readNumEntry("/mapeditor/stepsTotal",100);
	undoSet.setEntry ("/history/stepsTotal",QString::number(stepsTotal));
	
	// Initialize find routine
	itFind=NULL;				
	EOFind=false;

	printFrame=true;
	printFooter=true;

	blockReposition=false;
	blockSaveState=false;

	hidemode=HideNone;

	// Create temporary files
	makeTmpDirs();

	// Initially set movingCentre
	updateViewCenter();

	mapCenter->reposition();	//	for positioning heading

}

MapEditor::~MapEditor()
{
  if (imageBuffer) delete imageBuffer;
  if (urlOperator) {
    urlOperator->stop();
    delete urlOperator;
  }

	//cout <<"Destructor MapEditor\n";
}

MapCenterObj* MapEditor::getMapCenter()
{
    return mapCenter;
}

Q3Canvas* MapEditor::getCanvas()
{
    return mapCanvas;
}

void MapEditor::adjustCanvasSize()
{
	// To adjust the canvas to map, viewport size and position, we have to
	// do some coordinate magic...
	//
	// Get rectangle of (scroll-)view. 
	// We want to be in canvas coords, so
	// we map. Important if view is zoomed...
	QRect view = inverseWorldMatrix().mapRect( QRect( contentsX(), contentsY(),
												visibleWidth(), visibleHeight()) );	
												
	// Now we need the bounding box of view AND map to calc the correct canvas size.
	// Why? Because if the map itself is moved out of view, the view has to be enlarged
	// to avoid jumping aroung...
	QRect map=mapCenter->getTotalBBox();

	// right edge - left edge
	int cw= max(map.x() + map.width(),  view.x() + view.width())  - min(map.x(), view.x());
	int ch= max(map.y() + map.height(), view.y() + view.height()) - min(map.y(), view.y());


	if ( (cw!=mapCanvas->width()) || (ch!=mapCanvas->height()) ||
		!mapCanvas->onCanvas (map.topLeft()) || !mapCanvas->onCanvas (map.bottomRight())
	)	
	{	
		// move the map on canvas (in order to not move it on screen) this is neccessary
		// a) if topleft corner of canvas is left or above topleft corner of view and also left of
		//    above topleft corner of map. E.g. if map is completly inside view, but it would be possible 
		//    to scroll to an empty area of canvas to the left.
		// b) if topleft corner of map left of or above topleft of canvas
		int dx=0;
		int dy=0;

		if (cw > mapCanvas->width() )
		{
			if (map.x()<0) dx=-map.x();	
		}
		if (cw <  mapCanvas->width() )
			dx=-min (view.x(),map.x());
		if (ch > mapCanvas->height() )
		{
			if (map.y()<0) dy=-map.y();	
		}
		if (ch <  mapCanvas->height() )
		{
			dy=-min (view.y(),map.y());
		}
		// We really have to resize now. Let's go...
		mapCanvas->resize (cw,ch);
		if ( (dx!=0) || (dy!=0) ) 
		{
			mapCenter->moveAllBy(dx,dy);
			mapCenter->reposition();
//			mapCenter->positionBBox();	// To move float

			// scroll the view (in order to not move map on screen)
			scrollBy (dx,dy);
		}	
	}
}

bool MapEditor::isRepositionBlocked()
{
	return blockReposition;
}

QString MapEditor::getName (LinkableMapObj *lmo)
{
	QString s;
	if (!lmo) return QString("Error: NULL has no name!");

	if ((typeid(*lmo) == typeid(BranchObj) ||
				      typeid(*lmo) == typeid(MapCenterObj))) 
	{
		
		s=(((BranchObj*)lmo)->getHeading());
		if (s=="") s="unnamed";
		return QString("branch (%1)").arg(s);
		//return QString("branch (<font color=\"#0000ff\">%1</font>)").arg(s);
	}	
	if ((typeid(*lmo) == typeid(FloatImageObj) ))
		return QString ("floatimage [%1]").arg(((FloatImageObj*)lmo)->getOriginalFilename());
		//return QString ("floatimage [<font color=\"#0000ff\">%1</font>]").arg(((FloatImageObj*)lmo)->getOriginalFilename());
	return QString("Unknown type has no name!");
}

void MapEditor::makeTmpDirs()
{
	// Create unique temporary directories
	tmpMapDir=QDir::convertSeparators (tmpVymDir+QString("/mapeditor-%1").arg(mapNum));
	histPath=QDir::convertSeparators (tmpMapDir+"/history");
	QDir d;
	d.mkdir (tmpMapDir);
}

QString MapEditor::saveToDir(const QString &tmpdir, const QString &prefix, bool writeflags, const QPoint &offset, LinkableMapObj *saveSel)
{
	// tmpdir		temporary directory to which data will be written
	// prefix		mapname, which will be appended to images etc.
	// writeflags	Only write flags for "real" save of map, not undo
	// offset		offset of bbox of whole map in canvas. 
	//				Needed for XML export
	
	// Save Header
	QString ls;
	switch (linkstyle)
	{
		case StyleLine: 
			ls="StyleLine";
			break;
		case StyleParabel:
			ls="StyleParabel";
			break;
		case StylePolyLine:	
			ls="StylePolyLine";
			break;
		default:
			ls="StylePolyParabel";
			break;
	}	

	QString s="<?xml version=\"1.0\" encoding=\"utf-8\"?><!DOCTYPE vymmap>\n";
	QString colhint="";
	if (linkcolorhint==HeadingColor) 
		colhint=attribut("linkColorHint","HeadingColor");

	QString mapAttr=attribut("version",vymVersion);
	if (!saveSel || saveSel==mapCenter)
		mapAttr+= attribut("author",mapCenter->getAuthor()) +
				  attribut("comment",mapCenter->getComment()) +
			      attribut("date",mapCenter->getDate()) +
		          attribut("backgroundColor", mapCanvas->backgroundColor().name() ) +
		          attribut("linkStyle", ls ) +
		          attribut("linkColor", defLinkColor.name() ) +
		          attribut("defXLinkColor", defXLinkColor.name() ) +
		          attribut("defXLinkWidth", QString().setNum(defXLinkWidth,10) ) +
		          colhint; 
	s+=beginElement("vymmap",mapAttr);
	incIndent();

	// Find the used flags while traversing the tree
	standardFlagsDefault->resetUsedCounter();
	
	// Reset the counters before saving
	FloatImageObj (mapCanvas).resetSaveCounter();

	// Build xml recursivly
	if (!saveSel || typeid (*saveSel) == typeid (MapCenterObj))
		// Save complete map, if saveSel not set
		s+=mapCenter->saveToDir(tmpdir,prefix,writeflags,offset);
	else
	{
		if ( typeid(*saveSel) == typeid(BranchObj) )
			// Save Subtree
			s+=((BranchObj*)(saveSel))->saveToDir(tmpdir,prefix,offset);
		else if ( typeid(*saveSel) == typeid(FloatImageObj) )
			// Save image
			s+=((FloatImageObj*)(saveSel))->saveToDir(tmpdir,prefix);
			
		else if (selection && typeid(*selection)==typeid(BranchObj))
			// Save selected branch is saved from mainwindow		//FIXME maybe use "subtree" above?
			// FIXME is this possible at all? BO is already above...
			s+=((BranchObj*)selection)->saveToDir(tmpdir,prefix,offset);
	}

	// Save local settings
	s+=settings.getXMLData (destPath);

	// Save selection
	if (selection && !saveSel ) 
		s+=valueElement("select",selection->getSelectString());

	decIndent();
	s+=endElement("vymmap");

	if (writeflags)
		standardFlagsDefault->saveToDir (tmpdir+"/flags/","",writeflags);
	return s;
}

void MapEditor::saveStateChangingPart(LinkableMapObj *undoSel, LinkableMapObj* redoSel, const QString &rc, const QString &comment)
{
	// save the selected part of the map, Undo will replace part of map 
	QString undoSelection="";
	if (undoSel)
		undoSelection=undoSel->getSelectString();
	else
		qWarning ("MapEditor::saveStateChangingPart  no undoSel given!");
	QString redoSelection="";
	if (redoSel)
		redoSelection=undoSel->getSelectString();
	else
		qWarning ("MapEditor::saveStateChangingPart  no redoSel given!");
		

	saveState (PartOfMap,
		undoSelection, "addMapReplace (\"PATH\")",
		redoSelection, rc, 
		comment, 
		undoSel);
}

void MapEditor::saveStateRemovingPart(LinkableMapObj *redoSel, const QString &comment)
{
	if (!redoSel)
	{
		qWarning ("MapEditor::saveStateRemovingPart  no redoSel given!");
		return;
	}
	QString undoSelection=redoSel->getParObj()->getSelectString();
	QString redoSelection=redoSel->getSelectString();
	if (typeid(*redoSel) == typeid(BranchObj)  ) 
	{
		// save the selected branch of the map, Undo will insert part of map 
		saveState (PartOfMap,
			undoSelection, QString("addMapInsert (\"PATH\",%1)").arg(((BranchObj*)redoSel)->getNum()),
			redoSelection, "delete ()", 
			comment, 
			redoSel);
	}
}


void MapEditor::saveState(LinkableMapObj *undoSel, const QString &uc, LinkableMapObj *redoSel, const QString &rc, const QString &comment) 
{
	// "Normal" savestate: save commands, selections and comment
	// so just save commands for undo and redo
	// and use current selection

	QString redoSelection="";
	if (redoSel) redoSelection=redoSel->getSelectString();
	QString undoSelection="";
	if (undoSel) undoSelection=undoSel->getSelectString();

	saveState (UndoCommand,
		undoSelection, uc,
		redoSelection, rc, 
		comment, 
		NULL);
}

void MapEditor::saveState(const QString &undoSel, const QString &uc, const QString &redoSel, const QString &rc, const QString &comment) 
{
	// "Normal" savestate: save commands, selections and comment
	// so just save commands for undo and redo
	// and use current selection
	saveState (UndoCommand,
		undoSel, uc,
		redoSel, rc, 
		comment, 
		NULL);
}

		
void MapEditor::saveState(const SaveMode &savemode, const QString &undoSelection, const QString &undoCom, const QString &redoSelection, const QString &redoCom, const QString &comment, LinkableMapObj *saveSel)
{
	// Main saveState

	if (blockSaveState) return;

	/* TODO remove after testing
	*/
	cout << "ME::saveState()  "<<endl;
	
	int undosAvail=undoSet.readNumEntry ("/history/undosAvail",0);
	int redosAvail=undoSet.readNumEntry ("/history/redosAvail",0);
	int curStep=undoSet.readNumEntry ("/history/curStep",0);
	// Find out current undo directory
	if (undosAvail<stepsTotal) undosAvail++;
	curStep++;
	if (curStep>stepsTotal) curStep=1;
	
	QString backupXML="";
	QString bakMapName=QDir::convertSeparators (QString("history-%1").arg(curStep));
	QString bakMapDir=QDir::convertSeparators (tmpMapDir +"/"+bakMapName);
	QString bakMapPath=QDir::convertSeparators(bakMapDir+"/map.xml");

	// Create bakMapDir if not available
	QDir d(bakMapDir);
	if (!d.exists()) 
		makeSubDirs (bakMapDir);

	// Save depending on how much needs to be saved	
	if (saveSel)
		backupXML=saveToDir (bakMapDir,mapName+"-",false, QPoint (),saveSel);
		
	QString undoCommand="";
	if (savemode==UndoCommand)
	{
		undoCommand=undoCom;
	}	
	else if (savemode==PartOfMap )
	{
		undoCommand=undoCom;
		undoCommand.replace ("PATH",bakMapPath);
	}

	if (!backupXML.isEmpty())
		// Write XML Data to disk
		saveStringToDisk (QString(bakMapPath),backupXML);

	// We would have to save all actions in a tree, to keep track of 
	// possible redos after a action. Possible, but we are too lazy: forget about redos.
	redosAvail=0;

	// Write the current state to disk
	undoSet.setEntry ("/history/undosAvail",QString::number(undosAvail));
	undoSet.setEntry ("/history/redosAvail",QString::number(redosAvail));
	undoSet.setEntry ("/history/curStep",QString::number(curStep));
	undoSet.setEntry (QString("/history/step-%1/undoCommand").arg(curStep),undoCommand);
	undoSet.setEntry (QString("/history/step-%1/undoSelection").arg(curStep),undoSelection);
	undoSet.setEntry (QString("/history/step-%1/redoCommand").arg(curStep),redoCom);
	undoSet.setEntry (QString("/history/step-%1/redoSelection").arg(curStep),redoSelection);
	undoSet.setEntry (QString("/history/step-%1/comment").arg(curStep),comment);
	undoSet.setEntry (QString("/history/version"),vymVersion);
	undoSet.writeSettings(histPath);

	/* TODO remove after testing
	*/
	//cout << "          into="<< histPath.toStdString()<<endl;
	cout << "    stepsTotal="<<stepsTotal<<
	", undosAvail="<<undosAvail<<
	", redosAvail="<<redosAvail<<
	", curStep="<<curStep<<endl;
	cout << "    ---------------------------"<<endl;
	cout << "    comment="<<comment.toStdString()<<endl;
	cout << "    undoCom="<<undoCommand.toStdString()<<endl;
	cout << "    undoSel="<<undoSelection.toStdString()<<endl;
	cout << "    redoCom="<<redoCom.toStdString()<<endl;
	cout << "    redoSel="<<redoSelection.toStdString()<<endl;
	if (saveSel) cout << "    saveSel="<<saveSel->getSelectString().ascii()<<endl;
	cout << "    ---------------------------"<<endl;

	mainWindow->updateHistory (undoSet);
	setChanged();
	updateActions();
}

void MapEditor::parseAtom(const QString &atom)
{
	API api;
	QString s,t;
	int x,y;
	bool b,ok;

	// Split string s into command and parameters
	api.parseInput (atom);
	QString com=api.command();
	
	// External commands
	if (com=="addBranch")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else 
		{	
			QList <int> pl;
			pl << 0 <<1;
			if (api.checkParamCount(pl))
			{
				if (api.paramCount()==0)
					addNewBranchInt (-2);
				else
				{
					y=api.parInt (ok,0);
					if (ok ) addNewBranchInt (y);
				}
			}
		}
	} else if (com=="addBranchBefore")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else 
		{	
			if (api.paramCount()==0)
			{
				addNewBranchBefore ();
			}	
		}
	} else if (com==QString("addMapReplace"))
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{
			//s=api.parString (ok,0);	// selection
			t=api.parString (ok,0);	// path to map
			if (QDir::isRelativePath(t)) t=QDir::convertSeparators (tmpMapDir + "/"+t);
			addMapReplaceInt(selection->getSelectString(),t);	
		}
	} else if (com==QString("addMapInsert"))
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else 
		{	
			if (api.checkParamCount(2))
			{
				t=api.parString (ok,0);	// path to map
				y=api.parInt(ok,1);		// position
				if (QDir::isRelativePath(t)) t=QDir::convertSeparators (tmpMapDir + "/"+t);
				addMapInsertInt(t,y);	
			}
		}
	} else if (com=="colorItem")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( typeid(*selection) != typeid(BranchObj) && 
					typeid(*selection) != typeid(MapCenterObj))
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{	
			QColor c=api.parColor (ok,0);
			if (ok) colorItem (c);
		}	
	} else if (com=="colorBranch")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( typeid(*selection) != typeid(BranchObj) && 
					typeid(*selection) != typeid(MapCenterObj))
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{	
			QColor c=api.parColor (ok,0);
			if (ok) colorBranch (c);
		}	
	} else if (com=="cut")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( typeid(*selection) != typeid(BranchObj) && 
					typeid(*selection) != typeid(MapCenterObj) &&
					typeid(*selection) != typeid(FloatImageObj) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch or floatimage");
		} else if (api.checkParamCount(0))
		{	
			cut();
		}	
	} else if (com=="delete")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(0))
		{	
			deleteSelection();
		}	
	} else if (com=="deleteKeepChilds")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(0))
		{	
			deleteKeepChilds();
		}	
	} else if (com=="deleteChilds")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(0))
		{	
			deleteChilds();
		}	
	} else if (com=="linkBranchToPos")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(4))
		{
			// 0	selectstring of parent
			// 1	num in parent (for branches)
			// 2,3	x,y of mainbranch or mapcenter
			s=api.parString(ok,0);
			LinkableMapObj *dst=mapCenter->findObjBySelect (s);
			if (dst)
			{	
				if (typeid(*dst) == typeid(BranchObj) ) 
				{
					// Get number in parent
					x=api.parInt (ok,1);
					if (ok)
						((BranchObj*)selection)->moveBranchTo ((BranchObj*)(dst),x);
				} else if (typeid(*dst) == typeid(MapCenterObj) ) 
				{
					((BranchObj*)selection)->moveBranchTo ((BranchObj*)(dst),-1);
					// Get coordinates of mainbranch
					x=api.parInt (ok,2);
					if (ok)
					{
						y=api.parInt (ok,3);
						if (ok) ((BranchObj*)selection)->move (x,y);
					}
				}	
			}	
		}
	} else if (com=="moveBranchUp")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(0))
		{
			moveBranchUp();
		}	
	} else if (com=="moveBranchDown")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(0))
		{
			moveBranchDown();
		}	
	} else if (com=="move")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( typeid(*selection) != typeid(BranchObj) && 
					typeid(*selection) != typeid(MapCenterObj) &&
					typeid(*selection) != typeid(FloatImageObj) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch or floatimage");
		} else if (api.checkParamCount(2))
		{	
			x=api.parInt (ok,0);
			if (ok)
			{
				y=api.parInt (ok,1);
				if (ok) move (x,y);
			}
		}	
	} else if (com=="moveRel")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( typeid(*selection) != typeid(BranchObj) && 
					typeid(*selection) != typeid(MapCenterObj) &&
					typeid(*selection) != typeid(FloatImageObj) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch or floatimage");
		} else if (api.checkParamCount(2))
		{	
			x=api.parInt (ok,0);
			if (ok)
			{
				y=api.parInt (ok,1);
				if (ok) moveRel (x,y);
			}
		}	
	} else if (com=="paste")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( typeid(*selection) != typeid(BranchObj) && 
					typeid(*selection) != typeid(MapCenterObj) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(0))
		{	
			paste();
		}	
	} else if (com=="select")
	{
		if (api.checkParamCount(1))
		{
			s=api.parString(ok,0);
			if (ok) select (s);
		}	
	} else if (com=="setMapAuthor")
	{
		if (api.checkParamCount(1))
		{
			s=api.parString(ok,0);
			if (ok) setMapAuthor (s);
		}	
	} else if (com=="setMapComment")
	{
		if (api.checkParamCount(1))
		{
			s=api.parString(ok,0);
			if (ok) setMapComment(s);
		}	
	} else if (com=="setMapBackgroundColor")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{
			QColor c=api.parColor (ok,0);
			if (ok) setMapBackgroundColor (c);
		}	
	} else if (com=="setMapDefLinkColor")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{
			QColor c=api.parColor (ok,0);
			if (ok) setMapDefLinkColor (c);
		}	
	} else if (com=="setMapLinkStyle")
	{
		if (api.checkParamCount(1))
		{
			s=api.parString (ok,0);
			if (ok) setMapLinkStyle(s);
		}	
	} else if (com=="setHeading")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{
			s=api.parString (ok,0);
			if (ok) 
				setHeading (s);
		}	
	} else if (com=="setHideExport")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( typeid(*selection) != typeid(BranchObj) && 
					typeid(*selection) != typeid(FloatImageObj) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch or floatimage");
		} else if (api.checkParamCount(1))
		{
			b=api.parBool(ok,0);
			if (ok) setHideExport (b);
		}
	} else if (com=="setURL")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{
			s=api.parString (ok,0);
			if (ok) setURLInt(s);
		}	
	} else if (com=="setVymLink")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{
			s=api.parString (ok,0);
			if (ok) setVymLinkInt(s);
		}	
	}
	else if (com=="setFlag")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{
			s=api.parString(ok,0);
			if (ok) 
			{
				BranchObj* bo=(BranchObj*)selection;
				bo->activateStandardFlag(s);
				bo->updateFlagsToolbar();
			}	
		}
	}	
	else if (com=="unsetFlag")
	{
		if (!selection)
		{
			api.setError (Aborted,"Nothing selected");
		} else if ( (typeid(*selection) != typeid(BranchObj) && 
					 typeid(*selection) != typeid(MapCenterObj)) )
		{				  
			api.setError (Aborted,"Type of selection is not a branch");
		} else if (api.checkParamCount(1))
		{
			s=api.parString(ok,0);
			if (ok) 
			{
				BranchObj* bo=(BranchObj*)selection;
				bo->deactivateStandardFlag(s);
				bo->updateFlagsToolbar();
			}	
		}
	} else
	{
		api.setError (Aborted,"Unknown command");
	}

	// Any errors?
	if (api.errorLevel()==NoError)
		setChanged();
	else	
	{
		// TODO Error handling
		qWarning("MapEditor::parseAtom: Error!");
		qWarning(api.errorMessage());
	} 
}


bool MapEditor::isDefault()
{
    return mapDefault;
}

bool MapEditor::isUnsaved()
{
    return mapUnsaved;
}

bool MapEditor::hasChanged()
{
    return mapChanged;
}

void MapEditor::setChanged()
{
	mapChanged=true;
	mapDefault=false;
	mapUnsaved=true;
	findReset();
}

void MapEditor::closeMap()
{
	// Unselect before disabling the toolbar actions
	if (selection) selection->unselect();
	selection=NULL;
	updateActions();

    clear();
	close();
}

void MapEditor::setFilePath(QString fname)
{
	setFilePath (fname,fname);
}

void MapEditor::setFilePath(QString fname, QString destname)
{
	if (fname.isEmpty() || fname=="")
	{
		filePath="";
		fileName="";
		destPath="";
	} else
	{
		filePath=fname;		// becomes absolute path
		fileName=fname;		// gets stripped of path
		destPath=destname;	// needed for vymlinks

		// If fname is not an absolute path, complete it
		filePath=QDir(fname).absPath();
		fileDir=filePath.left (1+filePath.findRev ("/"));

		// Set short name, too. Search from behind:
		int i=fileName.findRev("/");
		if (i>=0) fileName=fileName.remove (0,i+1);

		// Forget the .vym (or .xml) for name of map
		mapName=fileName.left(fileName.findRev(".",-1,true) );
	}
}

QString MapEditor::getFilePath()
{
	return filePath;
}

QString MapEditor::getFileName()
{
	return fileName;
}

QString MapEditor::getMapName()
{
	return mapName;
}

QString MapEditor::getDestPath()
{
	return destPath;
}

ErrorCode MapEditor::load (QString fname, LoadMode lmode)
{
	ErrorCode err=success;

	if (lmode==NewMap)
	{
		if (selection) selection->unselect();
		selection=NULL;
		mapCenter->clear();
		mapCenter->setMapEditor(this);
		// (map state is set later at end of load...)
	} else
	{
		if (!selection || (typeid(*selection) != typeid(BranchObj) &&
			typeid(*selection) != typeid (MapCenterObj)))
			return aborted;
		BranchObj *bo=(BranchObj*)selection;	
		if (lmode==ImportAdd)
			saveStateChangingPart(
				selection,
				selection,
				QString("addMapInsert (%1)").arg(fname),
				QString("Add map %1 to %2").arg(fname).arg(getName(bo)));
		else	
			saveStateChangingPart(
				selection,
				selection,
				QString("addMapReplace(%1)").arg(fname),
				QString("Add map %1 to %2").arg(fname).arg(getName(bo)));
	}	
	
    
    mapBuilderHandler handler;
	QFile file( fname );

	// I am paranoid: file should exist anyway
	// according to check in mainwindow.
	if (!file.exists() )
	{
		QMessageBox::critical( 0, tr( "Critical Parse Error" ),
				   tr("Couldn't open map " +fname)+".");
		err=aborted;	
	} else
	{
		blockReposition=true;
		QXmlInputSource source( file);
		QXmlSimpleReader reader;
		reader.setContentHandler( &handler );
		reader.setErrorHandler( &handler );
		handler.setMapEditor( this );
		handler.setTmpDir (filePath.left(filePath.findRev("/",-1)));	// needed to load files with rel. path
		handler.setInputFile (file.name());
		handler.setLoadMode (lmode);
		blockSaveState=true;
		bool ok = reader.parse( source );
		blockReposition=false;
		blockSaveState=false;
		file.close();
		if ( ok ) 
		{
			mapCenter->reposition();
			adjustCanvasSize();
			if (lmode==NewMap)
			{
				mapDefault=false;
				mapChanged=false;
				mapUnsaved=false;
			}
		} else 
		{
			QMessageBox::critical( 0, tr( "Critical Parse Error" ),
					   tr( handler.errorProtocol() ) );
			// returnCode=1;	
			// Still return "success": the map maybe at least
			// partially read by the parser
		}	
	}	
	updateActions();
	return err;
}

int MapEditor::save (const SaveMode &savemode)
{
	int returnCode=0;

	// Create mapName and fileDir
	makeSubDirs (fileDir);
	QString fname;
	if (saveZipped())
		// save as .xml
		fname=mapName+".xml";
	else
		// use name given by user, even if he chooses .doc
		fname=fileName;


	QString saveFile;
	if (savemode==CompleteMap || selection==NULL)
		saveFile=saveToDir (fileDir,mapName+"-",true,QPoint(),NULL);
	else	
		saveFile=saveToDir (fileDir,mapName+"-",true,QPoint(),selection);

	if (!saveStringToDisk(fileDir+fname,saveFile))
		return 1;

	if (returnCode==0)
	{
		mapChanged=false;
		mapUnsaved=false;
		updateActions();
	}

	return returnCode;
}

void MapEditor::setZipped (bool z)
{
	zipped=z;
}

bool MapEditor::saveZipped ()
{
	return zipped;
}

void MapEditor::print()
{
	if ( !printer ) 
	{
		printer = new QPrinter;
		printer->setColorMode (QPrinter::Color);
		printer->setPrinterName (settings.value("/mainwindow/printerName",printer->printerName()).toString());
	}

	QRect totalBBox=mapCenter->getTotalBBox();

	// Try to set orientation automagically
	// Note: Interpretation of generated postscript is amibiguous, if 
	// there are problems with landscape mode, see
	// http://sdb.suse.de/de/sdb/html/jsmeix_print-cups-landscape-81.html

	if (totalBBox.width()>totalBBox.height())
		// recommend landscape
		printer->setOrientation (QPrinter::Landscape);
	else	
		// recommend portrait
		printer->setOrientation (QPrinter::Portrait);

	if ( printer->setup(this) ) 
	// returns false, if printing is canceled
	{
		QPainter pp(printer);

		// Don't print the visualisation of selection
		LinkableMapObj *oldselection=NULL;
		if (selection) 
		{
			oldselection=selection;
			selection->unselect();
		}

		// Handle sizes of map and paper:
		//
		// setWindow defines which part of the canvas will be transformed 
		// setViewport defines area on paper in device coordinates (dpi)
		// e.g. (0,50,700,700) is upper part on A4
		// see also /usr/lib/qt3/doc/html/coordsys.html

		Q3PaintDeviceMetrics metrics (printer);

		double paperAspect = (double)metrics.width()   / (double)metrics.height();
		double   mapAspect = (double)totalBBox.width() / (double)totalBBox.height();

		QRect mapRect=totalBBox;
		Q3CanvasRectangle *frame=NULL;
		Q3CanvasText *footerFN=NULL;
		Q3CanvasText *footerDate=NULL;
		if (printFrame || printFooter)
		{
			
			if (printFrame) 
			{
				// Print frame around map
				mapRect.setRect (totalBBox.x()-10, totalBBox.y()-10, 
					totalBBox.width()+20, totalBBox.height()+20);
				frame=new Q3CanvasRectangle (mapRect,mapCanvas);
				frame->setBrush (QColor(Qt::white));
				frame->setPen (QColor(Qt::black));
				frame->setZ(0);
				frame->show();    
			}		
			/* TODO remove after testing 
			QCanvasLine *l=new QCanvasLine (mapCanvas);
			l->setPoints (0,0,mapRect.width(),mapRect.height());
			l->setPen (QPen(QColor(black), 1));
			l->setZ (200);
			l->show();
			*/

			if (printFooter) 
			{
				// Print footer below map
				QFont font;		
				font.setPointSize(10);
				footerFN=new Q3CanvasText (mapCanvas);
				footerFN->setText ("VYM - " + fileName);
				footerFN->setFont(font);
				footerFN->move (mapRect.x(), mapRect.y() + mapRect.height() );
				footerFN->setZ(Z_TEXT);
				footerFN->show();    
				footerDate=new Q3CanvasText (mapCanvas);
				footerDate->setText (QDate::currentDate().toString(Qt::TextDate));
				footerDate->setFont(font);
				footerDate->move (mapRect.x()+mapRect.width()-footerDate->boundingRect().width(), mapRect.y() + mapRect.height() );
				footerDate->setZ(Z_TEXT);
				footerDate->show();    
			}
			pp.setWindow (mapRect.x(), mapRect.y(), mapRect.width(), mapRect.height()+20);
		}	else	
		{
			pp.setWindow (mapRect);
		}	

		if (mapAspect>=paperAspect)
		{
			// Fit horizontally to paper width
			pp.setViewport(0,0, metrics.width(),(int)(metrics.width()/mapAspect) );	
		}	else
		{
			// Fit vertically to paper height
			pp.setViewport(0,0,(int)(metrics.height()*mapAspect),metrics.height());	
		}	

		mapCanvas->drawArea(mapRect, &pp);	// draw Canvas to printer

		// Delete Frame and footer
		if (footerFN) 
		{
			delete (footerFN);
			delete (footerDate);
		}	
		if (frame)  delete (frame);

		// Restore selection
		if (oldselection) 
		{
			selection=oldselection;
			selection->select();
		}	

		// Save settings in vymrc
		settings.writeEntry("/mainwindow/printerName",printer->printerName());
	}
}

QPixmap MapEditor::getPixmap()
{
	QRect mapRect=mapCenter->getTotalBBox();
	QPixmap pix (mapRect.size());
	QPainter pp (&pix);

	// Don't print the visualisation of selection
	LinkableMapObj *oldselection=NULL;
	if (selection) 
	{
		oldselection=selection;
		selection->unselect();
	}

	pp.setWindow (mapRect);

	mapCanvas->drawArea(mapRect, &pp);	// draw Canvas to painter


	// Restore selection
	if (oldselection) 
	{
		selection=oldselection;
		selection->select();
	}	
	
	return pix;
}

void MapEditor::setHideTmpMode (HideTmpMode mode)
{
	hidemode=mode;
	mapCenter->setHideTmp (hidemode);
	mapCenter->reposition();
	adjustCanvasSize();
	canvas()->update();
}

HideTmpMode MapEditor::getHideTmpMode()
{
	return hidemode;
}

void MapEditor::exportImage(QString fn)
{
	setExportMode (true);
	QPixmap pix (getPixmap());
	pix.save(fn, "PNG");
	setExportMode (false);
}

void MapEditor::setExportMode (bool b)
{
	// should be called before and after exports
	// depending on the settings
	if (b && settings.value("/export/useHideExport","yes")=="yes")
		setHideTmpMode (HideExport);
	else	
		setHideTmpMode (HideNone);
}

void MapEditor::exportImage(QString fn, QString format)
{
	setExportMode (true);
	QPixmap pix (getPixmap());
	pix.save(fn, format);
	setExportMode (false);
}

void MapEditor::exportOOPresentation(const QString &fn, const QString &cf)
{
	ExportOO ex;
	ex.setFile (fn);
	ex.setMapCenter(mapCenter);
	if (ex.setConfigFile(cf)) 
	{
		setExportMode (true);
		ex.exportPresentation();
		setExportMode (false);
	}
}



void MapEditor::exportXML(const QString &dir)
{
	// Hide stuff during export, if settings want this
	setExportMode (true);

	// Create subdirectories
	makeSubDirs (dir);

	// write to directory
	QString saveFile=saveToDir (dir,mapName+"-",true,mapCenter->getTotalBBox().topLeft() ,NULL);
	QFile file;

	file.setName ( dir + "/"+mapName+".xml");
	if ( !file.open( QIODevice::WriteOnly ) )
	{
		// This should neverever happen
		QMessageBox::critical (0,tr("Critical Export Error"),tr("MapEditor::exportXML couldn't open %1").arg(file.name()));
		return;
	}	

	// Write it finally, and write in UTF8, no matter what 
	QTextStream ts( &file );
	ts.setEncoding (QTextStream::UnicodeUTF8);
	ts << saveFile;
	file.close();

	// Now write image, too
	exportImage (dir+"/images/"+mapName+".png");

	setExportMode (false);
}

void MapEditor::clear()
{
	if (selection)
	{
		selection->unselect();
		selection=NULL;
	}	

	mapCenter->clear();
}

void MapEditor::copy()
{
	if (selection) 
	{
		// write to directory
		QString clipfile="part";
		QString saveFile=saveToDir (fileDir,clipfile+"-",true,QPoint(),selection);
		QFile file;

		file.setName ( clipboardDir + "/"+clipfile+".xml");
		if ( !file.open( QIODevice::WriteOnly ) )
		{
			// This should neverever happen
			QMessageBox::critical (0,tr("Critical Export Error"),tr("MapEditor::exportXML couldn't open %1").arg(file.name()));
			return;
		}	

		// Write it finally, and write in UTF8, no matter what 
		QTextStream ts( &file );
		ts.setEncoding (QTextStream::UnicodeUTF8);
		ts << saveFile;
		file.close();

		clipboardEmpty=false;
		updateActions();
	}	    
}

void MapEditor::redo()
{
	blockSaveState=true;
	
	// Restore variables
	int curStep=undoSet.readNumEntry (QString("/history/curStep"));
	int undosAvail=undoSet.readNumEntry (QString("/history/undosAvail"));
	int redosAvail=undoSet.readNumEntry (QString("/history/redosAvail"));
	// Can we undo at all?
	if (redosAvail<1) return;
	redosAvail--;

	if (undosAvail<stepsTotal) undosAvail++;
	curStep++;
	if (curStep>stepsTotal) curStep=1;
	QString undoCommand=  undoSet.readEntry (QString("/history/step-%1/undoCommand").arg(curStep));
	QString undoSelection=undoSet.readEntry (QString("/history/step-%1/undoSelection").arg(curStep));
	QString redoCommand=  undoSet.readEntry (QString("/history/step-%1/redoCommand").arg(curStep));
	QString redoSelection=undoSet.readEntry (QString("/history/step-%1/redoSelection").arg(curStep));
	QString comment=undoSet.readEntry (QString("/history/step-%1/comment").arg(curStep));
	QString version=undoSet.readEntry ("/history/version");

	if (!checkVersion(version))
		QMessageBox::warning(0,tr("Warning"),
			tr("Version %1 of saved undo/redo data\ndoes not match current vym version %2.").arg(version).arg(vymVersion));


	// Find out current undo directory
	QString bakMapDir=QDir::convertSeparators (QString(tmpMapDir+"/undo-%1").arg(curStep));

/* TODO remove testing
*/
	cout << "ME::redo() begin\n";
	cout << "    undosAvail="<<undosAvail<<endl;
	cout << "    redosAvail="<<redosAvail<<endl;
	cout << "       curStep="<<curStep<<endl;
	cout << "    ---------------------------"<<endl;
	cout << "    comment="<<comment.toStdString()<<endl;
	cout << "    undoCom="<<undoCommand.toStdString()<<endl;
	cout << "    undoSel="<<undoSelection.toStdString()<<endl;
	cout << "    redoCom="<<redoCommand.toStdString()<<endl;
	cout << "    redoSel="<<redoSelection.toStdString()<<endl;
	cout << "    ---------------------------"<<endl<<endl;

	// select  object before redo
	if (!redoSelection.isEmpty())
		select (redoSelection);


	parseAtom (redoCommand);
	mapCenter->reposition();

	blockSaveState=false;

	undoSet.setEntry ("/history/undosAvail",QString::number(undosAvail));
	undoSet.setEntry ("/history/redosAvail",QString::number(redosAvail));
	undoSet.setEntry ("/history/curStep",QString::number(curStep));
	undoSet.writeSettings(histPath);

	mainWindow->updateHistory (undoSet);
	updateActions();

	/* TODO remove testing
	cout << "ME::redo() end\n";
	cout << "    undosAvail="<<undosAvail<<endl;
	cout << "    redosAvail="<<redosAvail<<endl;
	cout << "       curStep="<<curStep<<endl;
	cout << "    ---------------------------"<<endl<<endl;
	*/


}

bool MapEditor::isRedoAvailable()
{
	if (undoSet.readNumEntry("/history/redosAvail",0)>0)
		return true;
	else	
		return false;
}

void MapEditor::undo()
{
	blockSaveState=true;
	
	// Restore variables
	int curStep=undoSet.readNumEntry (QString("/history/curStep"));
	int undosAvail=undoSet.readNumEntry (QString("/history/undosAvail"));
	int redosAvail=undoSet.readNumEntry (QString("/history/redosAvail"));

	// Can we undo at all?
	if (undosAvail<1) return;

	QString undoCommand=  undoSet.readEntry (QString("/history/step-%1/undoCommand").arg(curStep));
	QString undoSelection=undoSet.readEntry (QString("/history/step-%1/undoSelection").arg(curStep));
	QString redoCommand=  undoSet.readEntry (QString("/history/step-%1/redoCommand").arg(curStep));
	QString redoSelection=undoSet.readEntry (QString("/history/step-%1/redoSelection").arg(curStep));
	QString comment=undoSet.readEntry (QString("/history/step-%1/comment").arg(curStep));
	QString version=undoSet.readEntry ("/history/version");

	if (!checkVersion(version))
		QMessageBox::warning(0,tr("Warning"),
			tr("Version %1 of saved undo/redo data\ndoes not match current vym version %2.").arg(version).arg(vymVersion));

	// Find out current undo directory
	QString bakMapDir=QDir::convertSeparators (QString(tmpMapDir+"/undo-%1").arg(curStep));

	// select  object before undo
	if (!undoSelection.isEmpty())
		select (undoSelection);

/* TODO testing
*/	
	cout << "ME::undo() begin\n";
	cout << "    undosAvail="<<undosAvail<<endl;
	cout << "    redosAvail="<<redosAvail<<endl;
	cout << "       curStep="<<curStep<<endl;
	cout << "    ---------------------------"<<endl;
	cout << "    comment="<<comment.toStdString()<<endl;
	cout << "    undoCom="<<undoCommand.toStdString()<<endl;
	cout << "    undoSel="<<undoSelection.toStdString()<<endl;
	cout << "    redoCom="<<redoCommand.toStdString()<<endl;
	cout << "    redoSel="<<redoSelection.toStdString()<<endl;
	cout << "    ---------------------------"<<endl<<endl;
	parseAtom (undoCommand);
	mapCenter->reposition();

	undosAvail--;
	curStep--; 
	if (curStep<1) curStep=stepsTotal;

	redosAvail++;

	blockSaveState=false;
/* TODO remove testing
	cout << "ME::undo() end\n";
	cout << "    undosAvail="<<undosAvail<<endl;
	cout << "    redosAvail="<<redosAvail<<endl;
	cout << "       curStep="<<curStep<<endl;
	cout << "    ---------------------------"<<endl<<endl;
*/

	undoSet.setEntry ("/history/undosAvail",QString::number(undosAvail));
	undoSet.setEntry ("/history/redosAvail",QString::number(redosAvail));
	undoSet.setEntry ("/history/curStep",QString::number(curStep));
	undoSet.writeSettings(histPath);

	mainWindow->updateHistory (undoSet);
	updateActions();
}

bool MapEditor::isUndoAvailable()
{
	if (undoSet.readNumEntry("/history/undosAvail",0)>0)
		return true;
	else	
		return false;
}

void MapEditor::gotoHistoryStep (int i)
{
	// Restore variables
	int undosAvail=undoSet.readNumEntry (QString("/history/undosAvail"));
	int redosAvail=undoSet.readNumEntry (QString("/history/redosAvail"));

	if (i<0) i=undosAvail+redosAvail;

	// Clicking above current step makes us undo things
	if (i<undosAvail) 
	{	
		for (int j=0; j<undosAvail-i; j++) undo();
		return;
	}	
	// Clicking below current step makes us redo things
	if (i>undosAvail) 
		for (int j=undosAvail; j<i; j++) redo();

	// And ignore clicking the current row ;-)	
}

void MapEditor::addMapReplaceInt(const QString &undoSel, const QString &path)
{
	QString pathDir=path.left(path.findRev("/"));
	QDir d(pathDir);
	QFile file (path);

	if (d.exists() )
	{
		// We need to parse saved XML data
		mapBuilderHandler handler;
		QXmlInputSource source( file);
		QXmlSimpleReader reader;
		reader.setContentHandler( &handler );
		reader.setErrorHandler( &handler );
		handler.setMapEditor( this );
		handler.setTmpDir ( pathDir );	// needed to load files with rel. path
		if (undoSel.isEmpty())
		{
			unselect();
			mapCenter->clear();
			handler.setLoadMode (NewMap);
		} else	
		{
			select (undoSel);
			handler.setLoadMode (ImportReplace);
		}	
		blockReposition=true;
		bool ok = reader.parse( source );
		blockReposition=false;
		if (! ok ) 
		{	
			// This should never ever happen
			QMessageBox::critical( 0, tr( "Critical Parse Error while reading %1").arg(path),
								    handler.errorProtocol());
		}
	} else	
		QMessageBox::critical( 0, tr( "Critical Error" ), tr("Could not read %1").arg(path));
}

void MapEditor::addMapInsertInt (const QString &path, int pos)
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
				      typeid(*selection) == typeid(MapCenterObj))) 
	{
		QString pathDir=path.left(path.findRev("/"));
		QDir d(pathDir);
		QFile file (path);

		if (d.exists() )
		{
			// We need to parse saved XML data
			mapBuilderHandler handler;
			QXmlInputSource source( file);
			QXmlSimpleReader reader;
			reader.setContentHandler( &handler );
			reader.setErrorHandler( &handler );
			handler.setMapEditor( this );
			handler.setTmpDir ( pathDir );	// needed to load files with rel. path
			handler.setLoadMode (ImportAdd);
			blockReposition=true;
			bool ok = reader.parse( source );
			blockReposition=false;
			if (! ok ) 
			{	
				// This should never ever happen
				QMessageBox::critical( 0, tr( "Critical Parse Error while reading %1").arg(path),
										handler.errorProtocol());
			}
			if (selection!=mapCenter)
				((BranchObj*)selection)->getLastBranch()->moveBranchTo ((BranchObj*)(selection),pos);
		} else	
			QMessageBox::critical( 0, tr( "Critical Error" ), tr("Could not read %1").arg(path));
	}		
}

void MapEditor::pasteNoSave()
{
	load (clipboardDir+"/part.xml",ImportAdd);
}

void MapEditor::cutNoSave()
{
	copy();
	deleteSelection();
}

void MapEditor::paste()
{   
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
				      typeid(*selection) == typeid(MapCenterObj))) 
	{
		pasteNoSave();
		saveStateChangingPart(
			selection,
			selection,
			"paste ()",
			QString("Paste to %1").arg( getName(selection))
		);
		mapCenter->reposition();
		adjustCanvasSize();
	}
}

void MapEditor::cut()
{
	saveStateChangingPart(
		selection->getParObj(),
		selection,
		"cut ()",
		QString("Cut %1").arg(getName(selection))
	);
	copy();
	cutNoSave();
	mapCenter->reposition();
	adjustCanvasSize();
}

void MapEditor::move(const int &x, const int &y)
{
	if (selection)
	{
		QString ps=qpointToString (selection->getAbsPos());
		QString s=selection->getSelectString();
		saveState(
			s, "move "+ps, 
			s, "move "+qpointToString (QPoint (x,y)), 
			QString("Move %1 to  %2").arg(getName(selection)).arg(ps));
		selection->move(x,y);
		mapCenter->reposition();
		adjustCanvasSize();
	}

}

void MapEditor::moveRel (const int &x, const int &y)
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
				      typeid(*selection) == typeid(MapCenterObj) ||
					  typeid(*selection) == typeid (FloatImageObj))) 
	{
		QString ps=qpointToString (selection->getRelPos());
		QString s=selection->getSelectString();
		saveState(
			s, "moveRel "+ps, 
			s, "moveRel "+qpointToString (QPoint (x,y)), 
			QString("Move %1 to relativ position %2").arg(getName(selection)).arg(ps));
		((OrnamentedObj*)selection)->move2RelPos (x,y);
		mapCenter->reposition();
		adjustCanvasSize();
	}
}

void MapEditor::moveBranchUp()
{
	BranchObj* bo;
	BranchObj* par;
	if (typeid(*selection) == typeid(BranchObj)  ) 
	{
		bo=(BranchObj*)selection;
		if (!bo->canMoveBranchUp()) return;
		par=(BranchObj*)(bo->getParObj());
		selection->unselect();
		bo=par->moveBranchUp (bo);	// bo will be the one below selection
		selection->select();
		saveState (selection,"moveBranchDown ()",bo,"moveBranchUp ()",QString("Move up %1").arg(getName(bo)));
		mapCenter->reposition();
		ensureSelectionVisible();
	}
}

void MapEditor::moveBranchDown()
{
	BranchObj* bo;
	BranchObj* par;
	if (typeid(*selection) == typeid(BranchObj)  ) 
	{
		bo=(BranchObj*)selection;
		if (!bo->canMoveBranchDown()) return;
		par=(BranchObj*)(bo->getParObj());
		selection->unselect(); 
		bo=par->moveBranchDown(bo);	// bo will be the one above selection
		selection->select();
		saveState(selection,"moveBranchUp ()",bo,"moveBranchDown ()",QString("Move down %1").arg(getName(bo)));
		mapCenter->reposition();
		ensureSelectionVisible();
	}	
}

QString MapEditor::getHeading(bool &ok, QPoint &p)
{
	if (selection  &&  
		 (typeid(*selection) == typeid(BranchObj) || 
		  typeid(*selection) == typeid(MapCenterObj) ) ) 
	{
		ok=true;
		ensureSelectionVisible();
		p = ((BranchObj*)selection)->getAbsPos();
		p.setX (p.x() - contentsX());
		p.setY (p.y() - contentsY() + ((BranchObj*)selection)->height()/2);
		return ((BranchObj*)selection)->getHeading();
	}
	ok=false;
	return QString();
}

void MapEditor::setHeading(const QString &s)
{
	if (selection  &&  
		 (typeid(*selection) == typeid(BranchObj) || 
		  typeid(*selection) == typeid(MapCenterObj) ) ) 
	{
		editingBO=(BranchObj*)selection;
		saveState(
			selection,
			"setHeading (\""+editingBO->getHeading()+"\")", 
			selection,
			"setHeading (\""+s+"\")", 
			QString("Set heading of %1 to \"%2\"").arg(getName(editingBO)).arg(s) );
		editingBO->setHeading(s );
		editingBO=NULL;
		mapCenter->reposition();
		adjustCanvasSize();
		ensureSelectionVisible();
	}
}

void MapEditor::setURLInt (const QString &s)
{
	// Internal function, no saveState needed
	if (selection  &&  
		 (typeid(*selection) == typeid(BranchObj) || 
		  typeid(*selection) == typeid(MapCenterObj) ) ) 
	{
		((BranchObj*)selection)->setURL(s);
		mapCenter->reposition();
		adjustCanvasSize();
		ensureSelectionVisible();
	}
}

void MapEditor::setHeadingInt(const QString &s)
{
	if (selection  &&  
		 (typeid(*selection) == typeid(BranchObj) || 
		  typeid(*selection) == typeid(MapCenterObj) ) ) 
	{
		((BranchObj*)selection)->setHeading(s);
		mapCenter->reposition();
		adjustCanvasSize();
		ensureSelectionVisible();
	}
}

void MapEditor::setVymLinkInt (const QString &s)
{
	// Internal function, no saveState needed
	if (selection  &&  
		 (typeid(*selection) == typeid(BranchObj) || 
		  typeid(*selection) == typeid(MapCenterObj) ) ) 
	{
		((BranchObj*)selection)->setVymLink(s);
		mapCenter->reposition();
		adjustCanvasSize();
		ensureSelectionVisible();
	}
}

BranchObj* MapEditor::addNewBranchInt(int num)
{
	// Depending on pos:
	// -3		insert in childs of parent  above selection 
	// -2		add branch to selection 
	// -1		insert in childs of parent below selection 
	// 0..n		insert in childs of parent at pos
	BranchObj *newbo=NULL;
	if (selection  &&  
		 (typeid(*selection) == typeid(BranchObj) || 
		  typeid(*selection) == typeid(MapCenterObj) ) ) 
	{
		BranchObj* bo = (BranchObj*) selection;
		if (num==-2)
		{
			// save scroll state. If scrolled, automatically select
			// new branch in order to tmp unscroll parent...
			return bo->addBranch();
			
		}else if (num==-1)
		{
			num=bo->getNum()+1;
			bo=(BranchObj*)bo->getParObj();
		}else if (num==-3)
		{
			num=bo->getNum();
			bo=(BranchObj*)bo->getParObj();
		}
		if (!bo) return bo;
		newbo=bo->insertBranch(num);
	}	
	return newbo;
}	

BranchObj* MapEditor::addNewBranch(int pos)
{
	// Different meaning than num in addNewBranchInt!
	// -1	add above
	//  0	add as child
	// +1	add below
	BranchObj *bo = (BranchObj*) selection;
	BranchObj *newbo=NULL;

	if (selection  &&  
		 (typeid(*selection) == typeid(BranchObj) || 
		  typeid(*selection) == typeid(MapCenterObj) ) ) 
	{
		newbo=addNewBranchInt (pos-2);

		if (newbo)
		{
			saveState(
				selection,		// FIXME sholdnt newbo be deleted here???
				"delete ()",
				selection,
				QString ("addBranch (%1)").arg(pos-2),
				QString ("Add new branch to %1").arg(getName(bo)));	

			mapCenter->reposition();
			adjustCanvasSize();
		}
	}	
	return newbo;
}


BranchObj* MapEditor::addNewBranchBefore()
{
	BranchObj *newbo=NULL;
	if (selection  &&  
		 (typeid(*selection) == typeid(BranchObj) ) )
		 // We accept no MapCenterObj here, so we _have_ a parent
	{
		BranchObj* bo = (BranchObj*) selection;
		QPoint p=bo->getRelPos();


		BranchObj *parbo=(BranchObj*)(selection->getParObj());

		// add below selection
		newbo=parbo->insertBranch(bo->getNum()+1);
		if (newbo)
		{
			newbo->move2RelPos (p);

			// Move selection to new branch
			((BranchObj*)selection)->moveBranchTo (newbo,-1);

			saveState (newbo, "deleteKeepChilds ()", newbo, "addBranchBefore ()", 
				QString ("Add branch before %1").arg(getName(bo)));

			mapCenter->reposition();
			adjustCanvasSize();
		}
	}	
	return newbo;
}

void MapEditor::deleteSelection()
{
	if (selection  && typeid(*selection) ==typeid(BranchObj) ) 
	{
		BranchObj* bo=(BranchObj*)selection;
		BranchObj* par=(BranchObj*)(bo->getParObj());
		bo->unselect();
		saveStateRemovingPart (bo, QString ("Delete %1").arg(getName(bo)));
		selection=NULL;
		par->removeBranch(bo);
		selection=par;
		selection->select();
		ensureSelectionVisible();
		mapCenter->reposition();
		adjustCanvasSize();
	}
	if (selection  && typeid(*selection) ==typeid(FloatImageObj) ) 
	{
		FloatImageObj* fio=(FloatImageObj*)selection;
		BranchObj* par=(BranchObj*)(fio->getParObj());
		saveStateChangingPart(
			par, 
			fio,
			"delete ()",
			QString("Delete %1").arg(getName(fio))
		);
		fio->unselect();
		selection=NULL;
		par->removeFloatImage(fio);
		selection=par;
		selection->select();
		ensureSelectionVisible();
		mapCenter->reposition();
		adjustCanvasSize();
	}
}

LinkableMapObj* MapEditor::getSelection()
{
	return selection;
}

void MapEditor::unselect()
{
	if (selection) 
	{
		selectionLast=selection;
		selection->unselect();
		selection=NULL;
	}
}	

void MapEditor::reselect()
{
	if (selectionLast)
	{
		selection=selectionLast;
		selection->select();
		selectionLast=NULL;
	}
}	

bool MapEditor::select (const QString &s)
{
	LinkableMapObj *lmo=mapCenter->findObjBySelect(s);

	// Finally select the found object
	if (lmo)
	{
		if (selection) unselect();
		selection=lmo;
		selection->select();
		adjustCanvasSize();
		ensureSelectionVisible();
		return true;
	} 
	return false;
}

QString MapEditor::getSelectString()
{
	if (selection) return selection->getSelectString();
	return QString();
}

void MapEditor::selectInt (LinkableMapObj *lmo)
{
	if (lmo && selection != lmo)
	{
		// select the MapObj
		if (selection) selection->unselect();
		selection=lmo;
		selection->select();
			
		adjustCanvasSize();
	}
}

void MapEditor::selectNextBranchInt()
{
	// Increase number of branch
	if (selection)
	{
		QString s=selection->getSelectString();
		QString part;
		QString typ;
		QString num;

		// Where am I? 
		part=s.section(",",-1);
		typ=part.left (3);
		num=part.right(part.length() - 3);

		s=s.left (s.length() -num.length());

		// Go to next lmo
		num=QString ("%1").arg(num.toUInt()+1);

		s=s+num;
		
		// Try to select this one
		if (select (s)) return;

		// We have no direct successor, 
		// try to increase the parental number in order to
		// find a successor with same depth

		int d=selection->getDepth();
		int oldDepth=d;
		int i;
		bool found=false;
		bool b;
		while (!found && d>0)
		{
			s=s.section (",",0,d-1);
			// replace substring of current depth in s with "1"
			part=s.section(",",-1);
			typ=part.left (3);
			num=part.right(part.length() - 3);

			if (d>1)
			{	
				// increase number of parent
				num=QString ("%1").arg(num.toUInt()+1);
				s=s.section (",",0,d-2) + ","+ typ+num;
			} else
			{
				// Special case, look at orientation
				if (selection->getOrientation()==OrientRightOfCenter)
					num=QString ("%1").arg(num.toUInt()+1);
				else	
					num=QString ("%1").arg(num.toUInt()-1);
				s=typ+num;
			}	

			if (select (s))
				// pad to oldDepth, select the first branch for each depth
				for (i=d;i<oldDepth;i++)
				{
					b=select (s);
					if (b)
					{	
						if ( ((BranchObj*)selection)->countBranches()>0)
							s+=",bo:0";
						else	
							break;
					} else
						break;
				}	

			// try to select the freshly built string
			found=select(s);
			d--;
		}
		return;
	}	
}

void MapEditor::selectPrevBranchInt()
{
	// Decrease number of branch
	if (selection)
	{
		QString s=selection->getSelectString();
		QString part;
		QString typ;
		QString num;

		// Where am I? 
		part=s.section(",",-1);
		typ=part.left (3);
		num=part.right(part.length() - 3);

		s=s.left (s.length() -num.length());

		// Go to next lmo
		num=QString ("%1").arg(num.toUInt()-1);

		s=s+num;
		
		// Try to select this one
		if (select (s)) return;

		// We have no direct precessor, 
		// try to decrease the parental number in order to
		// find a precessor with same depth

		int d=selection->getDepth();
		int oldDepth=d;
		int i;
		bool found=false;
		bool b;
		while (!found && d>0)
		{
			s=s.section (",",0,d-1);
			// replace substring of current depth in s with "1"
			part=s.section(",",-1);
			typ=part.left (3);
			num=part.right(part.length() - 3);

			if (d>1)
			{
				// decrease number of parent
				num=QString ("%1").arg(num.toUInt()-1);
				s=s.section (",",0,d-2) + ","+ typ+num;
			} else
			{
				// Special case, look at orientation
				if (selection->getOrientation()==OrientRightOfCenter)
					num=QString ("%1").arg(num.toUInt()-1);
				else	
					num=QString ("%1").arg(num.toUInt()+1);
				s=typ+num;
			}	

			if (select(s))
				// pad to oldDepth, select the last branch for each depth
				for (i=d;i<oldDepth;i++)
				{
					b=select (s);
					if (b)
						if ( ((BranchObj*)selection)->countBranches()>0)
							s+=",bo:"+ QString ("%1").arg( ((BranchObj*)selection)->countBranches()-1 );
						else	
							break;
					else
						break;
				}	
			
			// try to select the freshly built string
			found=select(s);
			d--;
		}
		return;
	}	
}

void MapEditor::selectUpperBranch()
{
	if (selection) 
	{
		if (typeid(*selection) == typeid(BranchObj))
		{
			if (selection->getOrientation()==OrientRightOfCenter)
				selectPrevBranchInt();
			else
				if (selection->getDepth()==1)
					selectNextBranchInt();
				else
					selectPrevBranchInt();
		}		
	}
}

void MapEditor::selectLowerBranch()
{
	if (selection) 
	{
		if (typeid(*selection) == typeid(BranchObj))
		{
			if (selection->getOrientation()==OrientRightOfCenter)
				selectNextBranchInt();
			else
				if (selection->getDepth()==1)
					selectPrevBranchInt();
				else
					selectNextBranchInt();
		}		
	}
}


void MapEditor::selectLeftBranch()
{
	BranchObj* bo;
	BranchObj* par;
	if (selection) 
	{
		if (typeid(*selection) == typeid(MapCenterObj))
		{
			par=  (BranchObj*) selection;
			bo=par->getLastSelectedBranch();
			if (bo)
			{
				// Workaround for reselecting on left and right side
				if (bo->getOrientation()==OrientRightOfCenter)
				{
					bo=par->getLastBranch();
				}	
				if (bo)
				{
					par->unselect();
					selection=bo;
					selection->select();
					adjustCanvasSize();
					ensureSelectionVisible();
				}
			}	
		} else
		{
			par=(BranchObj*)(selection->getParObj());
			if (selection->getOrientation()==OrientRightOfCenter)
			{
				if (typeid(*selection) == typeid(BranchObj) ||
					typeid(*selection) == typeid(FloatImageObj))
				{
					selection->unselect();
					selection=par;
					selection->select();
					adjustCanvasSize();
					ensureSelectionVisible();
				}
			} else
			{
				if (typeid(*selection) == typeid(BranchObj) )
				{
					bo=((BranchObj*)selection)->getLastSelectedBranch();
					if (bo) 
					{
						selection->unselect();
						selection=bo;
						selection->select();
						adjustCanvasSize();
						ensureSelectionVisible();
					}
				}
			}
		}	
	}
}

void MapEditor::selectRightBranch()
{
	BranchObj* bo;
	BranchObj* par;

	if (selection) 
	{
		if (typeid(*selection) == typeid(MapCenterObj))
		{
			par=  (BranchObj*) selection;
			bo=par->getLastSelectedBranch();
			if (bo)
			{
				// Workaround for reselecting on left and right side
				if (bo->getOrientation()==OrientLeftOfCenter)
					bo=par->getFirstBranch();
				if (bo)
				{
					par->unselect();
					selection=bo;
					selection->select();
					ensureSelectionVisible();
				}
			}
		} else
		{
			par=(BranchObj*)(selection->getParObj());
			if (selection->getOrientation()==OrientLeftOfCenter)
			{
				if (typeid(*selection) == typeid(BranchObj) ||
					typeid(*selection) == typeid(FloatImageObj))
				{
					selection->unselect();
					selection=par;
					selection->select();
					adjustCanvasSize();
					ensureSelectionVisible();
				}
			} else
			{
				if (typeid(*selection) == typeid(BranchObj) )
				{
					bo=((BranchObj*)selection)->getLastSelectedBranch();
					if (bo) 
					{
						selection->unselect();
						selection=bo;
						selection->select();
						adjustCanvasSize();
						ensureSelectionVisible();
					}
				}
			}
		}
	}
}

void MapEditor::selectFirstBranch()
{
	BranchObj *bo1;
	BranchObj *bo2;
	BranchObj* par;
	if (selection) {
		if (typeid(*selection) == typeid(BranchObj))
		{
			bo1=  (BranchObj*) selection;
			par=(BranchObj*)(bo1->getParObj());
			bo2=par->getFirstBranch();
			if (bo2) {
				bo1->unselect();
				selection=bo2;
				selection->select();
				ensureSelectionVisible();
			}
		}		
		adjustCanvasSize();
	}
}

void MapEditor::selectLastBranch()
{
	BranchObj *bo1;
	BranchObj *bo2;
	BranchObj* par;
	if (selection) {
		if (typeid(*selection) == typeid(BranchObj))
		{
			bo1=  (BranchObj*) selection;
			par=(BranchObj*)(bo1->getParObj());
			bo2=par->getLastBranch();
			if (bo2) {
				bo1->unselect();
				selection=bo2;
				selection->select();
				ensureSelectionVisible();
			}
		}		
		adjustCanvasSize();
	}
}

void MapEditor::selectMapBackgroundColor()
{
	QColor col = QColorDialog::getColor( mapCanvas->backgroundColor(), this );
	if ( !col.isValid() ) return;
	setBackgroundColor( col );
}


void MapEditor::setMapBackgroundColor(QColor col)
{
	QColor oldcol=mapCanvas->backgroundColor();
	saveState(
		selection,
		QString ("setMapBackgroundColor (%1)").arg(oldcol.name()),
		selection,
		QString ("setMapBackgroundColor (%1)").arg(col.name()),
		QString("Set background color of map to %1").arg(col.name()));
	mapCanvas->setBackgroundColor (col);
}

QColor MapEditor::getMapBackgroundColor()
{
    return mapCanvas->backgroundColor();
}

QColor MapEditor::getCurrentHeadingColor()
{
	if (selection) 
	{
		if (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj))
		{
			BranchObj *bo=(BranchObj*)selection;
			return bo->getColor(); 
		}    
	}
	
	QMessageBox::warning(0,tr("Warning"),tr("Can't get color of heading,\nthere's no branch selected"));
	return Qt::black;
}

void MapEditor::colorItem(QColor c) 
{
	if (selection) 
	{
		if (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj))
		{
			BranchObj *bo=(BranchObj*)selection;
			saveState(
				selection, 
				QString ("colorItem (%1)").arg(bo->getColor().name()),
				selection,
				QString ("colorItem (%1)").arg(c.name()),
				QString("Set color of %1 to %2").arg(getName(bo)).arg(c.name())
			);	
			bo->setColor(c); // color branch
		}    
	}
}

void MapEditor::colorBranch(QColor c)
{
	if (selection) 
	{
		if (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj))
		{
			BranchObj *bo=(BranchObj*)selection;
			saveStateChangingPart(
				selection, 
				selection,
				QString ("colorBranch (%1)").arg(c.name()),
				QString ("Set color of %1 and childs to %2").arg(getName(bo)).arg(c.name())
			);	
			bo->setColorChilds(c); // color links, color childs
		}    
	}
}


void MapEditor::toggleStandardFlag(QString f)
{
	if (selection)
	{
		if (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj))
		{
			BranchObj *bo=(BranchObj*)selection;
			QString u,r;
			if (bo->isSetStandardFlag(f))
			{
				r="unsetFlag";
				u="setFlag";
			}	
			else
			{
				u="unsetFlag";
				r="setFlag";
			}	
			saveState(
				selection,
				QString("%1 (\"%2\")").arg(u).arg(f), 
				selection,
				QString("%1 (\"%2\")").arg(r).arg(f),
				QString("Toggling standard flag \"%1\" of %2").arg(f).arg(getName(bo)));
			bo->toggleStandardFlag (f,mainWindow->useFlagGroups());
			adjustCanvasSize();
		}
	}	
}

void MapEditor::setViewCenter()
{
	// transform to CanvasView Coord:
	QPoint p=worldMatrix().map(movingCenter);
	center ( p.x(), p.y());
}


BranchObj* MapEditor::findText (QString s, bool cs)
{
	QTextDocument::FindFlags flags=0;
	if (cs) flags=QTextDocument::FindCaseSensitively;

	if (!itFind) 
	{	// Nothing found or new find process
		if (EOFind)
			// nothing found, start again
			EOFind=false;
		itFind=mapCenter->first();
	}	
	bool searching=true;
	bool foundNote=false;
	while (searching && !EOFind)
	{
		if (itFind)
		{
			// Searching in Note
			if (itFind->getNote().contains(s,cs))
			{
				if (selection!=itFind) 
				{
					if (selection) ((BranchObj*)selection)->unselect();
					selection=itFind;
					selection->select();
					adjustCanvasSize();
					ensureSelectionVisible();
				}
				if (textEditor->findText(s,flags)) 
				{
					searching=false;
					foundNote=true;
				}	
			}
			// Searching in Heading
			if (searching && itFind->getHeading().contains (s,cs) ) 
			{
				if (selection) ((BranchObj*)selection)->unselect();
				selection=itFind;
				selection->select();
				adjustCanvasSize();
				ensureSelectionVisible();
				searching=false;
			}
		}	
		if (!foundNote)
		{
			itFind=itFind->next();
			if (!itFind) EOFind=true;
		}
	}	
	if (!searching)
	{
		adjustCanvasSize();
		return (BranchObj*)selection;
	}	else
		return NULL;
}

void MapEditor::findReset()
{	// Necessary if text to find changes during a find process
	itFind=NULL;
	EOFind=false;
}
void MapEditor::setURL(const QString &url)
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
	{		
		BranchObj *bo=(BranchObj*)selection;
		QString oldurl=bo->getURL();
		bo->setURL (url);
		saveState (
			selection,
			QString ("setURL (\"%1\")").arg(oldurl),
			selection,
			QString ("setURL (\"%1\")").arg(url),
			QString ("set URL of %1 to %2").arg(getName(bo)).arg(url)
		);
		updateActions();
	}
}	

void MapEditor::editURL()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
	{		
		bool ok;
		BranchObj *bo=(BranchObj*)selection;
		QString text = QInputDialog::getText(
				"VYM", tr("Enter URL:"), QLineEdit::Normal,
				bo->getURL(), &ok, this );
		if ( ok) 
			// user entered something and pressed OK
			setURL (text);
	}
}

QString MapEditor::getURL()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
		return ((BranchObj*)selection)->getURL();
	else
		return "";
}

QStringList MapEditor::getURLs()
{
	QStringList urls;
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
	{		
		BranchObj *bo=(BranchObj*)selection;
		bo=bo->first();	
		while (bo) 
		{
			if (!bo->getURL().isEmpty()) urls.append( bo->getURL());
			bo=bo->next();
		}	
	}	
	return urls;
}


void MapEditor::editHeading2URL()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
		setURL (((BranchObj*)selection)->getHeading());
}	

void MapEditor::editBugzilla2URL()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
	{		
		BranchObj *bo=(BranchObj*)selection;
		QString url= "https://bugzilla.novell.com/show_bug.cgi?id="+bo->getHeading();
		setURL (url);
	}
}	

void MapEditor::editFATE2URL()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
	{		
		BranchObj *bo=(BranchObj*)selection;
		QString url= "http://keeper.suse.de:8080/webfate/match/id?value=ID"+bo->getHeading();
		saveState(
			selection,
			"setURL (\""+bo->getURL()+"\")",
			selection,
			"setURL (\""+url+"\")",
			QString("Use heading of %1 as link to FATE").arg(getName(bo))
		);	
		bo->setURL (url);
		updateActions();
	}
}	

void MapEditor::editVymLink()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
	{		
		BranchObj *bo=(BranchObj*)selection;
		QStringList filters;
		filters <<"VYM map (*.vym)";
		QFileDialog *fd=new QFileDialog( this,vymName+" - " +tr("Link to another map"));
		fd->setFilters (filters);
		fd->setCaption(vymName+" - " +tr("Link to another map"));
		if (! bo->getVymLink().isEmpty() )
			fd->selectFile( bo->getVymLink() );
		fd->show();

		QString fn;
		if ( fd->exec() == QDialog::Accepted )
		{
			saveState(
				selection,
				"setVymLink (\""+bo->getVymLink()+"\")",
				selection,
				"setVymLink (\""+fd->selectedFile()+"\")",
				QString("Set vymlink of %1 to %2").arg(getName(bo)).arg(fd->selectedFile())
			);	
			bo->setVymLink (fd->selectedFile() );
			updateActions();
			mapCenter->reposition();
			adjustCanvasSize();
			canvas()->update();
		}
	}
}

void MapEditor::deleteVymLink()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
	{		
		BranchObj *bo=(BranchObj*)selection;
		saveState(
			selection,
			"setVymLink (\""+bo->getVymLink()+"\")",
			selection,
			"setVymLink (\"\")",
			QString("Unset vymlink of %1").arg(getName(bo))
		);	
		bo->setVymLink ("" );
		updateActions();
		mapCenter->reposition();
		adjustCanvasSize();
		canvas()->update();
	}
}

void MapEditor::setHideExport(bool b)
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection)==typeid(FloatImageObj)))
	{
		OrnamentedObj *oo=(OrnamentedObj*)selection;
		oo->setHideInExport (b);
		QString u= b ? "false" : "true";
		QString r=!b ? "false" : "true";
		
		saveState(
			selection,
			QString ("setHideExport (%1)").arg(u),
			selection,
			QString ("setHideExport (%1)").arg(r),
			QString ("Set HideExport flag of %1 to %2").arg(getName(oo)).arg (r)
		);	
		updateActions();
		mapCenter->reposition();
		adjustCanvasSize();
		canvas()->update();
	}
}

void MapEditor::toggleHideExport()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection)==typeid(FloatImageObj)))
		setHideExport ( !((OrnamentedObj*)selection)->hideInExport() );
}

QString MapEditor::getVymLink()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
	{		
		return ((BranchObj*)selection)->getVymLink();
	}
	return "";
	
}

QStringList MapEditor::getVymLinks()
{
	QStringList links;
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
			typeid(*selection) == typeid(MapCenterObj)) )
	{		
		BranchObj *bo=(BranchObj*)selection;
		bo=bo->first();	
		while (bo) 
		{
			if (!bo->getVymLink().isEmpty()) links.append( bo->getVymLink());
			bo=bo->next();
		}	
	}	
	return links;
}


void MapEditor::deleteKeepChilds()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ))
	{		
		BranchObj* bo=(BranchObj*)selection;
		BranchObj* par=(BranchObj*)(bo->getParObj());
		QPoint p=bo->getRelPos();
		saveStateChangingPart(
			selection->getParObj(),
			selection,
			"deleteKeepChilds ()",
			QString("Remove %1 and keep its childs").arg(getName(bo))
		);

		QString sel=selection->getSelectString();
		unselect();
		par->removeBranchHere(bo);
		mapCenter->reposition();
		select (sel);
		((BranchObj*)selection)->move2RelPos (p);
		mapCenter->reposition();
		adjustCanvasSize();
	}	
}

void MapEditor::deleteChilds()
{
	if (selection && (typeid(*selection) == typeid(BranchObj) ||
		typeid(*selection)==typeid(MapCenterObj)))
	{		
		saveStateChangingPart(
			selection->getParObj(), 
			selection,
			"deleteChilds ()",
			QString( "Remove childs of branch %1").arg(getName(selection))
		);
		((BranchObj*)selection)->removeChilds();
		mapCenter->reposition();
	}	
}

void MapEditor::editMapInfo()
{
	ExtraInfoDialog dia;
	dia.setMapName (getFileName() );
	dia.setAuthor (mapCenter->getAuthor() );
	dia.setComment(mapCenter->getComment() );

	// Calc some stats
	QString stats;
    int i=0;
    Q3CanvasItemList l=canvas()->allItems();
    for (Q3CanvasItemList::Iterator it=l.begin(); it!=l.end(); ++it) 
        i++;
    stats+=QString ("%1 items on canvas\n").arg (i,6);

	uint b=0;
	uint f=0;
	uint n=0;
	uint xl=0;
	BranchObj *bo;
	bo=mapCenter->first();
	while (bo) 
	{
		if (!bo->getNote().isEmpty() ) n++;
		f+= bo->countFloatImages();
		b++;
		xl+=bo->countXLinks();
		bo=bo->next();
	}
    stats+=QString ("%1 branches\n").arg (b-1,6);
    stats+=QString ("%1 xLinks \n").arg (xl,6);
    stats+=QString ("%1 notes\n").arg (n,6);
    stats+=QString ("%1 images\n").arg (f,6);
	dia.setStats (stats);

	// Finally show dialog
	if (dia.exec() == QDialog::Accepted)
	{
		setMapAuthor (dia.getAuthor() );
		setMapComment (dia.getComment() );
	}
}

void MapEditor::updateActions()
{
	mainWindow->updateActions();
	// FIXME maybe don't update if blockReposition is set
}

void MapEditor::updateNoteFlag()
{
	if (selection)
		if ( (typeid(*selection) == typeid(BranchObj)) || 
			(typeid(*selection) == typeid(MapCenterObj))  )
			((BranchObj*)selection)->updateNoteFlag();
}

void MapEditor::setMapAuthor (const QString &s)
{
	saveState (
		selection,
		QString ("setMapAuthor (\"%1\")").arg(mapCenter->getAuthor()),
		selection,
		QString ("setMapAuthor (\"%1\")").arg(s),
		QString ("Set author of map to \"%1\"").arg(s)
	);
	mapCenter->setAuthor (s);
}

void MapEditor::setMapComment (const QString &s)
{
	saveState (
		selection,
		QString ("setMapComment (\"%1\")").arg(mapCenter->getComment()),
		selection,
		QString ("setMapComment (\"%1\")").arg(s),
		QString ("Set comment of map")
	);
	mapCenter->setComment (s);
}

void MapEditor::setMapLinkStyle (const QString & s)
{
	saveStateChangingPart (
		mapCenter,
		mapCenter,
		QString("setMapLinkStyle (\"%1\")").arg(s),
		QString("Set map link style (\"%1\")").arg(s)
	);	

	if (s=="StyleLine")
		linkstyle=StyleLine;
	else if (s=="StyleParabel")
		linkstyle=StyleParabel;
	else if (s=="StylePolyLine")
		linkstyle=StylePolyLine;
	else	
		linkstyle=StylePolyParabel;

	BranchObj *bo;
	bo=mapCenter->first();
	bo=bo->next();
	while (bo) 
	{
		bo->setLinkStyle(bo->getDefLinkStyle());
		bo=bo->next();
	}
	mapCenter->reposition();
}

LinkStyle MapEditor::getMapLinkStyle ()
{
	return linkstyle;
}	

void MapEditor::setMapDefLinkColor(QColor c)
{
	defLinkColor=c;
	updateActions();
}

void MapEditor::setMapLinkColorHintInt()
{
	// called from setMapLinkColorHint(lch) or at end of parse
	BranchObj *bo;
	bo=mapCenter->first();
	while (bo) 
	{
		bo->setLinkColor();
		bo=bo->next();
	}
}

void MapEditor::setMapLinkColorHint(LinkColorHint lch)
{
	linkcolorhint=lch;
	setMapLinkColorHintInt();
}

void MapEditor::toggleMapLinkColorHint()
{
	if (linkcolorhint==HeadingColor)
		linkcolorhint=DefaultColor;
	else	
		linkcolorhint=HeadingColor;
	BranchObj *bo;
	bo=mapCenter->first();
	while (bo) 
	{
		bo->setLinkColor();
		bo=bo->next();
	}
}

LinkColorHint MapEditor::getMapLinkColorHint()
{
	return linkcolorhint;
}

QColor MapEditor::getMapDefLinkColor()
{
	return defLinkColor;
}

void MapEditor::setMapDefXLinkColor(QColor col)
{
	defXLinkColor=col;
}

QColor MapEditor::getMapDefXLinkColor()
{
	return defXLinkColor;
}

void MapEditor::setMapDefXLinkWidth (int w)
{
	defXLinkWidth=w;
}

int MapEditor::getMapDefXLinkWidth()
{
	return defXLinkWidth;
}

void MapEditor::selectMapLinkColor()
{
	QColor col = QColorDialog::getColor( defLinkColor, this );
	if ( !col.isValid() ) return;
	saveState (
		selection,
		QString("setMapDefLinkColor (\"%1\")").arg(getMapDefLinkColor().name()),
		selection,
		QString("setMapDefLinkColor (\"%1\")").arg(col.name()),
		QString("Set link color to %1").arg(col.name())
	);
	setMapDefLinkColor( col );

}

void MapEditor::toggleScroll()
{
	if (selection && (typeid(*selection) == typeid(BranchObj)) )
	{
		BranchObj *bo=((BranchObj*)selection);
		if (bo->countBranches()==0) return;
		if (bo->getDepth()==0) return;
		QString u,r;
		if (bo->isScrolled())
		{
			r="unscroll";
			u="scroll";
		}	
		else	
		{
			u="scroll";
			r="unscroll";
		}	
		saveState(
			selection,
			QString ("%1 ()").arg(u),
			selection,
			QString ("%1 ()").arg(r),
			QString ("%1 %2").arg(r).arg(getName(bo))
		);
		bo->toggleScroll();
		adjustCanvasSize();
		canvas()->update();
	}
}

void MapEditor::unScrollAll()
{
	BranchObj *bo;
	bo=mapCenter->first();
	while (bo) 
	{
		if (bo->isScrolled()) bo->toggleScroll();
		bo=bo->next();
	}
}

void MapEditor::loadFloatImage ()
{
	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  )
	{
		BranchObj *bo=((BranchObj*)selection);

		Q3FileDialog *fd=new Q3FileDialog( this);
		fd->setMode (Q3FileDialog::ExistingFiles);
		fd->addFilter (QString (tr("Images") + " (*.png *.bmp *.xbm *.jpg *.png *.xpm *.gif *.pnm)"));
		ImagePreview *p =new ImagePreview (fd);
		fd->setContentsPreviewEnabled( TRUE );
		fd->setContentsPreview( p, p );
		fd->setPreviewMode( Q3FileDialog::Contents );
		fd->setCaption(vymName+" - " +tr("Load image"));
		fd->setDir (lastImageDir);
		fd->show();

		QString fn;
		if ( fd->exec() == QDialog::Accepted )
		{
			// FIXME in QT4 use:	lastImageDir=fd->directory();
			lastImageDir=QDir (fd->dirPath());
			QStringList flist = fd->selectedFiles();
			QStringList::Iterator it = flist.begin();
			FloatImageObj *fio;
			while( it != flist.end() ) 
			{
				fn = *it;
				bo->addFloatImage();
				fio=bo->getLastFloatImage();
				fio->load(*it);
				// FIXME check if load of fio was successful
				saveState(
					(LinkableMapObj*)fio,
					"delete ()",
					selection, 
					QString ("loadFloatImage (%1)").arg(*it),
					QString("Add floatimage %1 to %2").arg(*it).arg(getName(selection))
				);
				bo->getLastFloatImage()->setOriginalFilename(fn);
				++it;
			}

			mapCenter->reposition();
			adjustCanvasSize();
			canvas()->update();
		}
		delete (p);
		delete (fd);
	}
}

void MapEditor::saveFloatImage ()
{
	if (selection && 
		(typeid(*selection) == typeid(FloatImageObj)) )
	{
		FloatImageObj *fio=((FloatImageObj*)selection);
		QFileDialog *fd=new QFileDialog( this);
		fd->setFilters (imageIO.getFilters());
		fd->setCaption(vymName+" - " +tr("Save image"));
		fd->setFileMode( QFileDialog::AnyFile );
		fd->setDirectory (lastImageDir);
//		fd->setSelection (fio->getOriginalFilename());
		fd->show();

		QString fn;
		if ( fd->exec() == QDialog::Accepted )
		{
			if (QFile (fd->selectedFile()).exists() )
			{
				QMessageBox mb( vymName,
					tr("The file %1 exists already.\n"
					"Do you want to overwrite it?").arg(fd->selectedFile()),
				QMessageBox::Warning,
				QMessageBox::Yes | QMessageBox::Default,
				QMessageBox::Cancel | QMessageBox::Escape,
				QMessageBox::QMessageBox::NoButton );

				mb.setButtonText( QMessageBox::Yes, tr("Overwrite") );
				mb.setButtonText( QMessageBox::No, tr("Cancel"));
				switch( mb.exec() ) 
				{
					case QMessageBox::Yes:
						// save 
						break;;
					case QMessageBox::Cancel:
						// do nothing
						delete (fd);
						return;
						break;
				}
			}
			fio->save (fd->selectedFile(),imageIO.getType (fd->selectedFilter() ) );
		}
		delete (fd);
	}
}

void MapEditor::setFrame(const FrameType &t)
{
	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  )
	{
		selection->setFrameType (t);
		mapCenter->reposition();
		selection->updateLink();
	}
}

void MapEditor::setIncludeImagesVer(bool b)
{
	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  )
		((BranchObj*)selection)->setIncludeImagesVer(b);
		mapCenter->reposition();
}

void MapEditor::setIncludeImagesHor(bool b)
{
	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  )
		((BranchObj*)selection)->setIncludeImagesHor(b);
		mapCenter->reposition();
}

void MapEditor::setHideLinkUnselected (bool b)
{
	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  ||
		(typeid(*selection) == typeid(FloatImageObj)) )
		selection->setHideLinkUnselected(b);
}

void MapEditor::importDirInt(BranchObj *dst, QDir d)
{
	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  )
	{
		BranchObj *bo;
		
		// Traverse directories
		d.setFilter( QDir::Dirs| QDir::Hidden | QDir::NoSymLinks );
		QFileInfoList list = d.entryInfoList();
		QFileInfo fi;

		for (int i = 0; i < list.size(); ++i) 
		{
			fi=list.at(i);
			if (fi.fileName() != "." && fi.fileName() != ".." )
			{
				dst->addBranch();
				bo=dst->getLastBranch();
				bo->setHeading (fi.fileName() );
				bo->setColor (QColor("blue"));
				bo->toggleScroll();
				if ( !d.cd(fi.fileName()) ) 
					QMessageBox::critical (0,tr("Critical Import Error"),tr("Cannot find the directory %1").arg(fi.fileName()));
				else 
				{
					// Recursively add subdirs
					importDirInt (bo,d);
					d.cdUp();
				}
			}	
		}		
		// Traverse files
		d.setFilter( QDir::Files| QDir::Hidden | QDir::NoSymLinks );
		list = d.entryInfoList();

		for (int i = 0; i < list.size(); ++i) 
		{
			fi=list.at(i);
			dst->addBranch();
			bo=dst->getLastBranch();
			bo->setHeading (fi.fileName() );
			bo->setColor (QColor("black"));
			if (fi.fileName().right(4) == ".vym" )
				bo->setVymLink (fi.filePath());
		}	
	}		
}

void MapEditor::importDir()
{
	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  )
	{
		QStringList filters;
		filters <<"VYM map (*.vym)";
		QFileDialog *fd=new QFileDialog( this,vymName+ " - " +tr("Choose directory structure to import"));
		fd->setMode (QFileDialog::DirectoryOnly);
		fd->setFilters (filters);
		fd->setCaption(vymName+" - " +tr("Choose directory structure to import"));
		fd->show();

		QString fn;
		if ( fd->exec() == QDialog::Accepted )
		{
			BranchObj *bo=((BranchObj*)selection);
			importDirInt (bo,QDir(fd->selectedFile()) );
			mapCenter->reposition();
			adjustCanvasSize();
			canvas()->update();
		}
	}	
}

void MapEditor::followXLink(int i)
{
	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  )
	{
		BranchObj *bo=((BranchObj*)selection)->XLinkTargetAt(i);
		if (bo) 
		{
			selection->unselect();
			selection=bo;
			selection->select();
			ensureSelectionVisible();
		}
	}
}

void MapEditor::editXLink(int i)
{
	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  )
	{
		XLinkObj *xlo=((BranchObj*)selection)->XLinkAt(i);
		if (xlo) 
		{
			EditXLinkDialog dia;
			dia.setXLink (xlo);
			dia.setSelection(selection);
			if (dia.exec() == QDialog::Accepted)
			{
				if (dia.useSettingsGlobal() )
				{
					setMapDefXLinkColor (xlo->getColor() );
					setMapDefXLinkWidth (xlo->getWidth() );
				}
				if (dia.deleteXLink())
					((BranchObj*)selection)->deleteXLinkAt(i);
				//saveStateComplete("Edit xLink");	//FIXME undoCommand
			}
		}	
	}
}

void MapEditor::testFunction()
{
	cout << "MapEditor::testFunction() called\n";

	if (selection && 
		(typeid(*selection) == typeid(BranchObj)) || 
		(typeid(*selection) == typeid(MapCenterObj))  )
	{
		BranchObj* bo=(BranchObj*)selection;
		cout << bo->getHeading().ascii() <<" is scrolled: "<<bo->isScrolled()<<endl;
	}
	return;
	
	WarningDialog dia;
	dia.showCancelButton (true);
	dia.setText("This is a longer \nWarning");
	/*
	dia.setCaption("Warning: Flux problem");
	dia.setShowAgainName("/warnings/mapeditor");
	*/
	if (dia.exec()==QDialog::Accepted)
		cout << "accepted!\n";
	else	
		cout << "canceled!\n";
	return;

/* Hide hidden stuff temporary, maybe add this as regular function somewhere
	if (hidemode==HideNone)
	{
		setHideTmpMode (HideExport);
		mapCenter->calcBBoxSizeWithChilds();
		QRect totalBBox=mapCenter->getTotalBBox();
		QRect mapRect=totalBBox;
		QCanvasRectangle *frame=NULL;

		cout << "  map has =("<<totalBBox.x()<<","<<totalBBox.y()<<","<<totalBBox.width()<<","<<totalBBox.height()<<")\n";
	
		mapRect.setRect (totalBBox.x(), totalBBox.y(), 
			totalBBox.width(), totalBBox.height());
		frame=new QCanvasRectangle (mapRect,mapCanvas);
		frame->setBrush (QColor(white));
		frame->setPen (QColor(black));
		frame->setZ(0);
		frame->show();    
	}	
	else	
	{
		setHideTmpMode (HideNone);
	}	
	cout <<"  hidemode="<<hidemode<<endl;
	*/
}

void MapEditor::ensureSelectionVisible()
{
	if (selection)
	{
		LinkableMapObj* lmo= dynamic_cast <LinkableMapObj*> (selection);
		QPoint p;
		if (selection->getOrientation() == OrientLeftOfCenter)
			p= worldMatrix().map(QPoint (lmo->x(),lmo->y()));
		else	
			p= worldMatrix().map(QPoint (lmo->x()+lmo->width(),lmo->y()+lmo->height()));
		ensureVisible (p.x(), p.y() );
	}

}

void MapEditor::updateViewCenter()
{
	// Update movingCenter, so that we can zoom comfortably later
	QRect rc = QRect( contentsX(), contentsY(),
				  visibleWidth(), visibleHeight() );
	QRect canvasRect = inverseWorldMatrix().mapRect(rc);
	movingCenter.setX((canvasRect.right() + canvasRect.left())/2);
	movingCenter.setY((canvasRect.top() + canvasRect.bottom())/2);
}

void MapEditor::contentsContextMenuEvent ( QContextMenuEvent * e )
{
	// Lineedits are already closed by preceding
	// mouseEvent, we don't need to close here.

    QPoint p = inverseWorldMatrix().map(e->pos());
    LinkableMapObj* lmo=mapCenter->findMapObj(p, NULL);
	
    if (lmo) 
	{	// MapObj was found
		if (selection != lmo)
		{
			// select the MapObj
			if (selection) selection->unselect();
			selection=lmo;
			selection->select();
			adjustCanvasSize();
		}
		// Context Menu 
		if (selection) 
		{
			if (typeid(*selection)==typeid(BranchObj) ||
				typeid(*selection)==typeid(MapCenterObj) )
			{
				// Context Menu on branch or mapcenter
				updateActions();
				branchContextMenu->popup(e->globalPos() );
			}	
			if (typeid(*selection)==typeid(FloatImageObj))
			{
				// Context Menu on floatimage
				updateActions();
				floatimageContextMenu->popup(e->globalPos() );
			}	
		}	
	} else 
	{ // No MapObj found, we are on the Canvas itself
		// Context Menu on Canvas
		updateActions();
		canvasContextMenu->popup(e->globalPos() );
    } 
	e->accept();
}

void MapEditor::keyPressEvent(QKeyEvent* e)
{
	if (e->modifiers() && Qt::ControlModifier)
	{
		switch (mainWindow->getModMode())
		{
			case ModModeColor: 
				setCursor (PickColorCursor);
				break;
			case ModModeCopy: 
				setCursor (CopyCursor);
				break;
			case ModModeXLink: 
				setCursor (XLinkCursor);
				break;
			default :
				setCursor (Qt::ArrowCursor);
				break;
		} 
	}	
}

void MapEditor::keyReleaseEvent(QKeyEvent* e)
{
	if (!(e->modifiers() && Qt::ControlModifier))
		setCursor (Qt::ArrowCursor);
}

void MapEditor::contentsMousePressEvent(QMouseEvent* e)
{
	// Ignore right clicks, these will go to context menus
	if (e->button() == Qt::RightButton )
	{
		e->ignore();
		return;
	}

    QPoint p = inverseWorldMatrix().map(e->pos());
    LinkableMapObj* lmo=mapCenter->findMapObj(p, NULL);
	
	e->accept();

	//Take care of clickdesystem flags _or_ modifier modes
	//
	if (lmo && (typeid(*lmo)==typeid(BranchObj) ||
		typeid(*lmo)==typeid(MapCenterObj) ))
	{
		QString foname=((BranchObj*)lmo)->getSystemFlagName(p);
		if (!foname.isEmpty())
		{
			// systemFlag clicked
			selectInt (lmo);
			if (foname=="url") 
			{
				if (e->state() & Qt::ControlModifier)
					mainWindow->editOpenURLTab();
				else	
					mainWindow->editOpenURL();
			}	
			else if (foname=="vymLink")
			{
				mainWindow->editOpenVymLink();
				// tabWidget may change, better return now
				// before segfaulting...
			} else if (foname=="note")
				mainWindow->windowToggleNoteEditor();
			else if (foname=="hideInExport")		
				toggleHideExport();
			return;	
		} 
	} 
	// No system flag clicked, take care of modmodes

	// Special case: CTRL is pressed
	if (e->state() & Qt::ControlModifier)
	{
		if (mainWindow->getModMode()==ModModeColor)
		{
				pickingColor=true;
				setCursor (PickColorCursor);
				return;
		} 
		if (mainWindow->getModMode()==ModModeXLink)
		{	
			BranchObj *bo_begin=NULL;
			if (lmo)
				bo_begin=(BranchObj*)(lmo);
			else	
				if (selection && 
					((typeid(*selection) == typeid(BranchObj)) || 
					(typeid(*selection) == typeid(MapCenterObj)))  )
				bo_begin=(BranchObj*)selection;
			if (bo_begin)	
			{
				drawingLink=true;
				linkingObj_src=bo_begin;
				tmpXLink=new XLinkObj (mapCanvas);
				tmpXLink->setBegin (bo_begin);
				tmpXLink->setEnd   (p);
				tmpXLink->setColor(defXLinkColor);
				tmpXLink->setWidth(defXLinkWidth);
				tmpXLink->updateXLink();
				tmpXLink->setVisibility (true);
				return;
			} 
		}
	}
    if (lmo) 
	{	
		selectInt (lmo);
		// Left Button	    Move Branches
		if (e->button() == Qt::LeftButton )
		{
			movingObj_start.setX( p.x() - selection->x() );	
			movingObj_start.setY( p.y() - selection->y() );	
			movingObj_orgPos.setX (lmo->x() );
			movingObj_orgPos.setY (lmo->y() );
			movingObj_orgRelPos=lmo->getRelPos();

			// If modMode==copy, then we want to "move" the _new_ object around
			// then we need the offset from p to the _old_ selection, because of tmp
			if (mainWindow->getModMode()==ModModeCopy &&
				e->state() & Qt::ControlModifier)
			{
				if (typeid(*selection)==typeid(BranchObj) )
				{
					copyingObj=true;
					mapCenter->addBranch ((BranchObj*)selection);
					unselect();
					selection=mapCenter->getLastBranch();
					selection->select();
					mapCenter->reposition();
				}
			}	
			movingObj=selection;	
		} else
			// Middle Button    Toggle Scroll
			// (On Mac OS X this won't work, but we still have 
			// a button in the toolbar)
			if (e->button() == Qt::MidButton )
				toggleScroll();
		updateActions();
	} else 
	{ // No MapObj found, we are on the Canvas itself
		// Left Button	    move Pos of CanvasView
		if (e->button() == Qt::LeftButton )
		{
			movingObj=NULL;	// move Content not Obj
			movingObj_start=e->globalPos();
			movingCont_start=QPoint (contentsX(), contentsY() );
			movingVec=QPoint(0,0);
			setCursor(HandOpenCursor);
		} 
    } 
}

void MapEditor::contentsMouseMoveEvent(QMouseEvent* e)
{
	QPoint p = inverseWorldMatrix().map(e->pos());

    // Move the selected MapObj
    if ( selection && movingObj) 
    {	
		// To avoid jumping of the CanvasView, only 
		// ensureSelectionVisible, if not tmp linked
		if (!selection->hasParObjTmp())
			ensureSelectionVisible ();
		
		// Now move the selection, but add relative position 
		// (movingObj_start) where selection was chosen with 
		// mousepointer. (This avoids flickering resp. jumping 
		// of selection back to absPos)
		
		LinkableMapObj *lmosel;
		lmosel =  dynamic_cast <LinkableMapObj*> (selection);

		// Check if we could link 
		LinkableMapObj* lmo=mapCenter->findMapObj(p, lmosel);
		

		if (typeid(*selection) == typeid(FloatImageObj))
		{
			FloatObj *fo=(FloatObj*)selection;
			fo->move   (p.x() -movingObj_start.x(), p.y()-movingObj_start.y() );		
			fo->setRelPos();
			fo->reposition();

			// Relink float to new mapcenter or branch, if shift is pressed	
			// Only relink, if selection really has a new parent
			if ( (e->modifiers()==Qt::ShiftModifier) && lmo &&
				( (typeid(*lmo)==typeid(BranchObj)) ||
				  (typeid(*lmo)==typeid(MapCenterObj)) ) &&
				( lmo != fo->getParObj())  
				)
			{
				if (typeid(*fo) == typeid(FloatImageObj)) 
				{
					//TODO undocom
					//saveStateComplete(QString("Relink %1 to %2").arg(getName(fo)).arg(getName(lmo) ) );
					FloatImageObj *fio=(FloatImageObj*)(fo);
					((BranchObj*)(lmo))->addFloatImage (fio);
					fio->unselect();
					((BranchObj*)(fio->getParObj()))->removeFloatImage (fio);
					fio=((BranchObj*)(lmo))->getLastFloatImage();
					fio->setRelPos();
					fio->reposition();
					selection=(LinkableMapObj*)(fio);
					selection->select();
					movingObj=(MapObj*)(fio);
				}	
			}
		} else	
		{	// selection != a FloatObj
			if (lmosel->getDepth()==0)
			{
				// Move MapCenter
				if (e->buttons()== Qt::LeftButton && e->modifiers()==Qt::ShiftModifier) 
					mapCenter->moveAll(p.x() -movingObj_start.x(), p.y()-movingObj_start.y() );		
				else	
					mapCenter->move   (p.x() -movingObj_start.x(), p.y()-movingObj_start.y() );		
				mapCenter->updateRelPositions();	
			} else
			{	
				if (lmosel->getDepth()==1)
				{
					// Move mainbranch
					lmosel->move(p.x() -movingObj_start.x(), p.y()-movingObj_start.y() );		
					lmosel->setRelPos();
				} else
				{
					// Move ordinary branch
					if (lmosel->getOrientation() == OrientLeftOfCenter)
						// Add width of bbox here, otherwise alignRelTo will cause jumping around
						lmosel->move(p.x() -movingObj_start.x()+lmosel->getBBox().width(), 
							p.y()-movingObj_start.y() +lmosel->getTopPad() );		
					else	
						lmosel->move(p.x() -movingObj_start.x(), p.y()-movingObj_start.y() -lmosel->getTopPad());
				} 
				// reposition subbranch
				lmosel->reposition();	

				if (lmo && (lmo!=selection) &&  
					(typeid(*lmo) == typeid(BranchObj) ||
					(typeid(*lmo) == typeid(MapCenterObj) )
					) )
				{
					if (e->modifiers()==Qt::ControlModifier)
					{
						// Special case: CTRL to link below lmo
						lmosel->setParObjTmp (lmo,p,+1);
					}
					else if (e->modifiers()==Qt::ShiftModifier)
						lmosel->setParObjTmp (lmo,p,-1);
					else
						lmosel->setParObjTmp (lmo,p,0);
				} else	
				{
					lmosel->unsetParObjTmp();
				}		
			} // depth>0

		} // no FloatImageObj

		canvas()->update();
		return;
	} // selection && moving_obj
		
	// Draw a link from one branch to another
	if (drawingLink)
	{
		 tmpXLink->setEnd (p);
		 tmpXLink->updateXLink();
	}	 
	
    // Move CanvasView 
    if (!movingObj && !pickingColor &&!drawingLink) 
	{
		QPoint p=e->globalPos();
		movingVec.setX(-p.x() + movingObj_start.x() );
		movingVec.setY(-p.y() + movingObj_start.y() );
		setContentsPos( movingCont_start.x() + movingVec.x(),
	    movingCont_start.y() + movingVec.y());

		updateViewCenter();
    }
}


void MapEditor::contentsMouseReleaseEvent(QMouseEvent* e)
{
	LinkableMapObj *dst;
	// Have we been picking color?
	if (pickingColor)
	{
		pickingColor=false;
		setCursor (Qt::ArrowCursor);
		// Check if we are over another branch
		dst=mapCenter->findMapObj(inverseWorldMatrix().map(e->pos() ), NULL);
		if (dst && selection) 
		{	
			if (e->state() & Qt::ShiftModifier)
			{
				((BranchObj*)selection)->setColor (((BranchObj*)(dst))->getColor());
				((BranchObj*)selection)->setLinkColor ();
			}	
			else	
			{
				((BranchObj*)selection)->setColorChilds (((BranchObj*)(dst))->getColor());
				((BranchObj*)selection)->setLinkColor ();
			}	
		} 
		return;
	}

	// Have we been drawing a link?
	if (drawingLink)
	{
		drawingLink=false;
		// Check if we are over another branch
		dst=mapCenter->findMapObj(inverseWorldMatrix().map(e->pos() ), NULL);
		if (dst && selection) 
		{	
			tmpXLink->setEnd ( ((BranchObj*)(dst)) );
			tmpXLink->updateXLink();
			tmpXLink->activate();
			//saveStateComplete(QString("Activate xLink from %1 to %2").arg(getName(tmpXLink->getBegin())).arg(getName(tmpXLink->getEnd())) );	//FIXME undoCommand
		} else
		{
			delete(tmpXLink);
			tmpXLink=NULL;
		}
		return;
	}
	
    // Have we been moving something?
    if ( selection && movingObj ) 
    {	
		if(typeid(*selection)==typeid (FloatImageObj))
		{
			// Moved FloatObj. Maybe we need to reposition
			FloatImageObj *fo=(FloatImageObj*)selection;
		    QString pold=qpointToString(movingObj_orgRelPos);
		    QString pnow=qpointToString(fo->getRelPos());
			saveState(
				selection,
				"moveRel "+pold,
				selection,
				"moveRel "+pnow,
				QString("Move %1 to relativ position %2").arg(getName(selection)).arg(pnow));

			// FIXME Why calling parObj here?
			selection->getParObj()->requestReposition();
			mapCenter->reposition();
		}	

		// Check if we are over another branch, but ignore 
		// any found LMOs, which are FloatObjs
		dst=mapCenter->findMapObj(inverseWorldMatrix().map(e->pos() ), 
			((LinkableMapObj*)selection) );

		if (dst && (typeid(*dst)!=typeid(BranchObj) && typeid(*dst)!=typeid(MapCenterObj))) 
			dst=NULL;
		
		if (typeid(*selection) == typeid(MapCenterObj)  )
		{	// The MapCenter was moved
			cout << "FIXME missing savestate...\n";	
		}
		
		if (typeid(*selection) == typeid(BranchObj)  )
		{	// A branch was moved
			
			// save the position in case we link to mapcenter
			QPoint savePos=QPoint (selection->x(),selection->y() );

			// Reset the temporary drawn link to the original one
			((LinkableMapObj*)selection)->unsetParObjTmp();

			// For Redo we may need to save original selection
			QString preSelStr=selection->getSelectString();

			copyingObj=false;	
			if (dst ) 
			{
				BranchObj* bsel=(BranchObj*)selection;
				BranchObj* bdst=(BranchObj*)dst;

				QString preParStr=(bsel->getParObj())->getSelectString();
				QString preNum=QString::number (bsel->getNum(),10);
				QString preDstParStr;

				if (e->state() & Qt::ShiftModifier && dst->getParObj())
				{	// Link above dst
					preDstParStr=dst->getParObj()->getSelectString();
					bsel->moveBranchTo ( (BranchObj*)(bdst->getParObj()), bdst->getNum());
				} else 
				if (e->state() & Qt::ControlModifier && dst->getParObj())
				{
					// Link below dst
					preDstParStr=dst->getParObj()->getSelectString();
					bsel->moveBranchTo ( (BranchObj*)(bdst->getParObj()), bdst->getNum()+1);
				} else	
				{	// Append to dst
					preDstParStr=dst->getSelectString();
					bsel->moveBranchTo (bdst,-1);
					if (dst->getDepth()==0) bsel->move (savePos);
				} 
				QString postSelStr=selection->getSelectString();
				QString postNum=QString::number (bsel->getNum(),10);

				QString undoCom="linkBranchToPos (\""+ 
					preParStr+ "\"," + preNum  +"," + 
					QString ("%1,%2").arg(movingObj_orgPos.x()).arg(movingObj_orgPos.y())+ ")";

				QString redoCom="linkBranchToPos (\""+ 
					preDstParStr + "\"," + postNum + "," +
					QString ("%1,%2").arg(savePos.x()).arg(savePos.y())+ ")";

				saveState (
					postSelStr,undoCom,
					preSelStr, redoCom,
					QString("Relink %1 to %2").arg(getName(bsel)).arg(getName(dst)) );
			} else
				if (selection->getDepth()==1)
				{
					// The select string might be different _after_ moving around.
					// Therefor reposition and then use string of old selection, too
					mapCenter->reposition();

					QString ps=qpointToString ( ((BranchObj*)selection)->getRelPos() );
					saveState(
						selection->getSelectString(), "moveRel "+qpointToString(movingObj_orgRelPos), 
						preSelStr, "moveRel "+ps, 
						QString("Move %1 to relative position %2").arg(getName(selection)).arg(ps));
			
				}
			// Draw the original link, before selection was moved around
			mapCenter->reposition();
		}
		// Finally resize canvas, if needed
		adjustCanvasSize();
		canvas()->update();
		movingObj=NULL;		

		// Just make sure, that actions are still ok,e.g. the move branch up/down buttons...
		updateActions();
	} else 
		// maybe we moved View: set old cursor
		setCursor (Qt::ArrowCursor);
    
}

void MapEditor::contentsMouseDoubleClickEvent(QMouseEvent* e)
{
	if (e->button() == Qt::LeftButton )
	{
		QPoint p = inverseWorldMatrix().map(e->pos());
		LinkableMapObj *lmo=mapCenter->findMapObj(p, NULL);
		if (lmo) {	// MapObj was found
			// First select the MapObj than edit heading
			if (selection) selection->unselect();
			selection=lmo;
			selection->select();
			mainWindow->editHeading();
		}
	}
}

void MapEditor::resizeEvent (QResizeEvent* e)
{
	Q3CanvasView::resizeEvent( e );
	adjustCanvasSize();
}

void MapEditor::contentsDragEnterEvent(QDragEnterEvent *event) 
{

//  for (unsigned int i=0;event->format(i);i++) // Debug mime type
//    cerr << event->format(i) << endl;

  if (selection && 
      (typeid(*selection) == typeid(BranchObj)) || 
      (typeid(*selection) == typeid(MapCenterObj))) {
    
    // If QImageDrag can decode mime type 
    if (Q3ImageDrag::canDecode(event)) {
      event->accept();
      return;
    }
    
    // If image are dragged from firefox 
    if (event->provides("application/x-moz-file-promise-url") && 
	event->provides("application/x-moz-nativeimage")) {
      event->accept(true);
      return;
    }

    // If QUriDrag can decode mime type 
    if (Q3UriDrag::canDecode(event)) {
      event->accept();
      return;
    }
    
	// If Uri are dragged from firefox 
    if (event->provides("_NETSCAPE_URL")){
      event->accept();
      return;
    }

    // If QTextDrag can decode mime type
    if (Q3TextDrag::canDecode(event)) {
      event->accept();
      return;
    }

  }
  event->ignore();
}

bool isUnicode16(const QByteArray &d) 
{
  // TODO: make more precise check for unicode 16.
  // Guess unicode16 if any of second bytes are zero
  unsigned int length = max(0,d.size()-2)/2;
  for (unsigned int i = 0; i<length ; i++)
    if (d.at(i*2+1)==0) return true;
  return false;
}
      
void MapEditor::contentsDropEvent(QDropEvent *event) 
{
	if (selection && 
      (typeid(*selection) == typeid(BranchObj)) || 
      (typeid(*selection) == typeid(MapCenterObj))) 
	{
		bool update=false;
		Q3StrList uris;
		QString heading;
		if (event->provides("image/png")) 
		{
			QPixmap pix;
			if (Q3ImageDrag::decode(event, pix)) 
			{
				addFloatImageInt(pix);
				event->accept();
				update=true;
			} else
				event->ignore();

		} else if (event->provides("application/x-moz-file-promise-url") && 
			 event->provides("application/x-moz-nativeimage")) 
		{
			// Contains url to the img src in unicode16
			QByteArray d = event->encodedData("application/x-moz-file-promise-url");
			QString url = QString((const QChar*)d.data(),d.size()/2);
			fetchImage(url);
			event->accept();
			update=true;
		} else if (event->provides ("text/uri-list"))
		{	// Uris provided e.g. by konqueror
			Q3UriDrag::decode (event,uris);
		} else if (event->provides ("_NETSCAPE_URL"))
		{	// Uris provided by Mozilla
		  QStringList l = QStringList::split("\n", event->encodedData("_NETSCAPE_URL"));
		  uris.append(l[0]);
		  heading = l[1];
		} else if (event->provides("text/html")) {

		  // Handels text mime types
		  // Look like firefox allways handle text as unicode16 (2 bytes per char.)
		  QByteArray d = event->encodedData("text/html");
		  QString text;
		  if (isUnicode16(d)) 
		    text = QString((const QChar*)d.data(),d.size()/2);
		  else 
		    text = QString(d);

		  textEditor->setText(text);

		  event->accept();
		  update=true;
		} else if (event->provides("text/plain")) {
		  QByteArray d = event->encodedData("text/plain");
		  QString text;
		  if (isUnicode16(d))
		    text = QString((const QChar*)d.data(),d.size()/2);
		  else 
		    text = QString(d);

		  textEditor->setText(text);
		  
		  event->accept();
		  update= true;
		}

		if (uris.count()>0)
		{
			QStringList files;
			QStringList urls;
			QString s;
			BranchObj *bo;
			for (const char* u=uris.first(); u; u=uris.next())
			{
				bo=((BranchObj*)selection)->addBranch();
				if (bo)
				{
					s=Q3UriDrag::uriToLocalFile(u);
					if (!s.isEmpty()) 
					{
                       QString file = QDir::convertSeparators(s);
                       heading = QFileInfo(file).baseName();
                       files.append(file);
                       if (file.endsWith(".vym", false))
                           bo->setVymLink(file);
                       else
                           bo->setURL(u);
                   } else 
				   {
                       urls.append (u);
                       bo->setURL(u);
                   }

                   if (!heading.isEmpty())
                       bo->setHeading(heading);
                   else
                       bo->setHeading(u);
				}
			}
			update=true;
		}

		if (update) 
		{
			//FIXME saveState has to be called earlier for each of the drops...
			//saveStateComplete("Drop Event");	//TODO undo Command
			mapCenter->reposition();
			adjustCanvasSize();
			canvas()->update();
		}	
	}	
}

void MapEditor::addFloatImageInt (const QPixmap &img) 
{
  if (selection && 
      (typeid(*selection) == typeid(BranchObj)) || 
      (typeid(*selection) == typeid(MapCenterObj))  )
  {
    BranchObj *bo=((BranchObj*)selection);
    //FIXME XXX saveStateChangingPart(selection,QString("Add floatimage to %1").arg(getName(bo)));
    //QString fn=fd->selectedFile();
    //lastImageDir=fn.left(fn.findRev ("/"));
	FloatImageObj *fio=bo->addFloatImage();
    fio->load(img);
    fio->setOriginalFilename("Image added by Drag and Drop");
    mapCenter->reposition();
    adjustCanvasSize();
    canvas()->update();
  }
}


void MapEditor::imageDataFetched(const QByteArray &a, Q3NetworkOperation */*nop*/) 
{
  if (!imageBuffer) imageBuffer = new QBuffer();
  if (!imageBuffer->isOpen()) {
    imageBuffer->open(QIODevice::WriteOnly | QIODevice::Append);
  }
  imageBuffer->at(imageBuffer->at()+imageBuffer->writeBlock(a));
}


void MapEditor::imageDataFinished(Q3NetworkOperation *nop) 
{
	if (nop->state()==Q3NetworkProtocol::StDone) {
		QPixmap img(imageBuffer->buffer());
		addFloatImageInt (img);
	}

	if (imageBuffer) {
		imageBuffer->close();
		if (imageBuffer) {
			imageBuffer->close();
			delete imageBuffer;
			imageBuffer = 0;
		}
	}
}

void MapEditor::fetchImage(const QString &url) 
{
  if (urlOperator) {
    urlOperator->stop();
    disconnect(urlOperator);
    delete urlOperator;
  }
  
  urlOperator = new Q3UrlOperator(url);
  connect(urlOperator, SIGNAL(finished(Q3NetworkOperation *)), 
	  this, SLOT(imageDataFinished(Q3NetworkOperation*)));

  connect(urlOperator, SIGNAL(data(const QByteArray &, Q3NetworkOperation *)),
	  this, SLOT(imageDataFetched(const QByteArray &, Q3NetworkOperation *)));
  urlOperator->get();
}
